
from epyk.core.html import geo


class Plotly(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "Plotly"

  def scattermapbox(self, record, lon_columns=None, lat_columns=None, text_columns=None, title=None, filters=None,
                    profile=None, options=None,  width=(100, "%"), height=(430, "px"), htmlCode=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Scatter`

    Related Pages:

      https://plot.ly/javascript/mapbox-layers/

    :param record:
    :param lon_columns:
    :param lat_columns:
    :param text_columns:
    :param title:
    :param filters:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    series = []
    for i, l in enumerate(lon_columns):
      series.append({'lon': [], 'lat': [], 'text': []})
      for rec in record:
        series[-1]['lon'].append(rec.get(l, 0))
        series[-1]['lat'].append(rec.get(lat_columns[i], 0))
        if text_columns is not None:
          series[-1]['text'].append(rec.get(text_columns[i], 0))
    line_chart = geo.GeoPlotly.Scatter(self.parent.context.rptObj, width, height, title, options or {}, htmlCode,
                                        filters, profile)
    line_chart.options.responsive = True
    self.parent.context.register(line_chart)
    for i, s in enumerate(series):
      line_chart.add_trace(s)
      line_chart.data.marker.color = self.parent.context.rptObj.theme.colors[::-1][i]
    line_chart.layout.mapbox.style = "open-street-map"
    # line_chart.data.marker.size = 4
    # line_chart.layout.mapbox.style = "open-street-map"
    # line_chart.layout.mapbox.center.lat = 38
    # line_chart.layout.mapbox.center.lon = -90
    # line_chart.layout.dragmode = 'zoom'
    # line_chart.layout.mapbox.zoom = 3
    return line_chart

  def density(self, record, y_columns=None, x_axis=None, title=None, filters=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Scatter`

    Related Pages:

      https://plot.ly/javascript/mapbox-density-heatmaps/

    :param record:
    :param y_columns:
    :param x_axis:
    :param title:
    :param filters:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    line_chart = geo.GeoPlotly.Scatter(self.parent.context.rptObj, width, height, title, options or {}, htmlCode,
                                       filters, profile)
    line_chart.options.responsive = True
    self.parent.context.register(line_chart)
    line_chart.add_trace({"lon": [-112.8352], 'lat': [48.4113], 'z': [20], 'text': [0.0875]}, type="densitymapbox")
    line_chart.data.marker.color = "fuchsia"
    line_chart.data.marker.size = 4
    line_chart.layout.mapbox.style = "stamen-terrain"
    line_chart.layout.mapbox.center.lat = 38
    line_chart.layout.mapbox.center.lon = -90
    line_chart.layout.dragmode = 'zoom'
    line_chart.layout.zoom = 3
    return line_chart

  def chorolet(self, record, y_columns=None, x_axis=None, title=None, filters=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Chorolet`

    Related Pages:

      https://plotly.com/javascript/mapbox-county-choropleth/

    :param record:
    :param y_columns:
    :param x_axis:
    :param title:
    :param filters:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    data = {"locations": ["NY", "MA", "VT"], "z": [-50, -10, -20]}
    line_chart = geo.GeoPlotly.Chorolet(self.parent.context.rptObj, width, height, title, options or {}, htmlCode,
                                       filters, profile)
    line_chart.options.responsive = True
    self.parent.context.register(line_chart)
    data['geojson'] = "https://raw.githubusercontent.com/python-visualization/folium/master/examples/data/us-states.json"
    line_chart.add_trace(data)
    line_chart.layout.mapbox.center.lat = 38
    line_chart.layout.mapbox.center.lon = -90
    line_chart.layout.dragmode = 'zoom'
    line_chart.layout.zoom = 3
    return line_chart


class PlotlyBubble(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "Plotly"

  def bubble(self, scope, record=None, y_column=None, x_axis=None, title=None, profile=None, options=None,
             width=(100, "%"), height=(430, "px"), htmlCode=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.BubbleGeo`

    Related Pages:

      https://plotly.com/javascript/bubble-maps/

    :param scope:
    :param record:
    :param y_column:
    :param x_axis:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    map_chart = geo.GeoPlotly.BubbleGeo(self.parent.context.rptObj, width, height, title, options or {}, htmlCode,
                                         profile)
    map_chart.options.responsive = True
    data = {}
    if record is None:
      record = []
    for rec in record:
      if x_axis in rec:
        data[rec[x_axis]] = data.get(x_axis, 0) + float(rec.get(y_column, 0))
    self.parent.context.register(map_chart)
    if record:
      locations = list(data.keys())
      values = [data[k] for k in locations]
      map_chart.add_trace({'locations': locations})
      map_chart.data.marker.colorbar.title = "Test"
      map_chart.data.marker.line.color = "black"
      map_chart.data.marker.size = values
      # map_chart.data.marker.cmin = 0
      # map_chart.data.marker.cmax = max(values)
      map_chart.data.marker.color = values
      map_chart.data.marker.colorscale = 'Reds'
    map_chart.layout.geo.scope = scope
    map_chart.layout.geo.resolution = 150
    if width[1] == 'px':
      map_chart.layout.width = width[0]
    return map_chart

  def usa(self, record=None, lon_columns=None, lat_columns=None, text_columns=None, title=None, profile=None, options=None,
             width=(100, "%"), height=(430, "px"), htmlCode=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.BubbleGeo`

    Related Pages:

      https://plotly.com/javascript/bubble-maps/

    :param record:
    :param lon_columns:
    :param lat_columns:
    :param text_columns:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    map_usa = self.bubble('usa', record, None, None, title, profile, options, width, height, htmlCode)

    series = []
    for i, l in enumerate(lon_columns):
      series.append({'lon': [], 'lat': [], 'text': []})
      for rec in record:
        series[-1]['lon'].append(rec.get(l, 0))
        series[-1]['lat'].append(rec.get(lat_columns[i], 0))
        if text_columns is not None:
          series[-1]['text'].append(rec.get(text_columns[i], 0))
    for i, s in enumerate(series):
      map_usa.add_trace(s)
    map_usa.data.locationmode = 'USA-states'
    map_usa.layout.no_background()
    return map_usa

  def europe(self, record=None, y_column=None, x_axis=None, title=None, profile=None, options=None,
             width=(100, "%"), height=(430, "px"), htmlCode=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.BubbleGeo`

    Related Pages:

      https://plotly.com/javascript/bubble-maps/

    :param record:
    :param y_column:
    :param x_axis:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    return self.bubble('europe', record, y_column, x_axis, title, profile, options, width, height, htmlCode)


class PlotlyScatter(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "Plotly"

  def scatter(self, scope, record=None, lon_columns=None, lat_columns=None, text_columns=None, title="", profile=None, options=None,
              width=(100, "%"), height=(430, "px"), htmlCode=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.ScatterGeo`

    Related Pages:

      https://plotly.com/javascript/scatter-plots-on-maps/

    :param scope:
    :param record:
    :param lon_columns:
    :param lat_columns:
    :param text_columns:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    map_chart = geo.GeoPlotly.ScatterGeo(self.parent.context.rptObj, width, height, title, options or {}, htmlCode,
                                         profile)
    map_chart.options.responsive = True

    series = []
    for i, l in enumerate(lon_columns):
      series.append({'lon': [], 'lat': [], 'text': []})
      for rec in record:
        series[-1]['lon'].append(rec.get(l, 0))
        series[-1]['lat'].append(rec.get(lat_columns[i], 0))
        if text_columns is not None:
          series[-1]['text'].append(rec.get(text_columns[i], 0))

    self.parent.context.register(map_chart)
    for i, s in enumerate(series):
      map_chart.add_trace(s)
      map_chart.data.marker.colorbar.title = title
      map_chart.data.marker.line.color = self.parent.context.rptObj.theme.colors[-i]
      #map_chart.data.marker.size = values
      #map_chart.data.marker.cmin = 0
      #map_chart.data.marker.cmax = max(values)
      #map_chart.data.marker.color = values
      map_chart.data.marker.colorscale = 'Reds'
    map_chart.layout.geo.scope = scope
    if width[1] == 'px':
      map_chart.layout.width = width[0]
    return map_chart

  def north_america(self, record=None, lon_columns=None, lat_columns=None, text_columns=None, title=None, profile=None,
                    options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.ScatterGeo`

    Related Pages:

      https://plotly.com/javascript/scatter-plots-on-maps/

    :param record:
    :param lon_columns:
    :param lat_columns:
    :param text_columns:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    map_chart = self.scatter('north america', record, lon_columns, lat_columns, text_columns, title, profile, options, width, height, htmlCode)
    return map_chart

  def usa(self, record=None, lon_columns=None, lat_columns=None, text_columns=None, title=None, profile=None,
          options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.ScatterGeo`

    Related Pages:

      https://plotly.com/javascript/scatter-plots-on-maps/

    :param record:
    :param lon_columns:
    :param lat_columns:
    :param text_columns:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    map_chart = self.scatter('usa', record, lon_columns, lat_columns, text_columns, title, profile, options, width, height, htmlCode)
    return map_chart


class PlotlyChoropleth(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "Plotly"

  def world_map(self, record, y_columns=None, country_column=None, title=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Choropleth`

    Related Pages:

      https://plotly.com/javascript/choropleth-maps/

    :param record:
    :param y_columns:
    :param country_column:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    series = []
    for y in y_columns:
      s = {'locations': [], 'z': []}
      for rec in record:
        s['locations'].append(rec[country_column])
        s['z'].append(rec[y])
      series.append(s)
    line_chart = geo.GeoPlotly.Choropleth(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    line_chart.options.responsive = True
    self.parent.context.register(line_chart)
    for s in series:
      line_chart.add_trace(s)
    line_chart.data.locationmode = 'country names'
    return line_chart

  def europe(self, record, y_column=None, x_axis=None, title=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Choropleth`

    Related Pages:

      https://plotly.com/javascript/choropleth-maps/

    :param record:
    :param y_column:
    :param x_axis:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    map_chart = geo.GeoPlotly.Choropleth(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    map_chart.options.responsive = True
    data = {}
    for rec in record:
      if x_axis in rec:
        data[rec[x_axis]] = data.get(x_axis, 0) + float(rec.get(y_column, 0))
    self.parent.context.register(map_chart)
    locations = list(data.keys())
    map_chart.add_trace({'locations': locations, 'z': [data[k] for k in locations]})
    map_chart.data.locationmode = 'country names'
    map_chart.data.autocolorscale = True
    map_chart.layout.geo.scope = 'europe'
    map_chart.layout.geo.showlakes = True
    map_chart.layout.geo.showland = True
    return map_chart

  def asia(self, record, y_column=None, x_axis=None, title=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Choropleth`

    Related Pages:

      https://plotly.com/javascript/choropleth-maps/

    :param record:
    :param y_column:
    :param x_axis:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    map_chart = geo.GeoPlotly.Choropleth(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    map_chart.options.responsive = True
    data = {}
    for rec in record:
      if x_axis in rec:
        data[rec[x_axis]] = data.get(x_axis, 0) + float(rec.get(y_column, 0))

    map_chart.options.responsive = True
    self.parent.context.register(map_chart)
    locations = list(data.keys())
    map_chart.add_trace({'locations': locations, 'z': [data[k] for k in locations]})
    map_chart.data.locationmode = 'country names'
    map_chart.layout.geo.scope = 'asia'
    map_chart.layout.geo.showlakes = True
    map_chart.layout.geo.showland = True
    return map_chart

  def africa(self, record, y_column=None, x_axis=None, title=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Choropleth`

    Related Pages:

      https://plotly.com/javascript/choropleth-maps/

    :param record:
    :param y_column:
    :param x_axis:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    map_chart = geo.GeoPlotly.Choropleth(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    map_chart.options.responsive = True
    data = {}
    for rec in record:
      if x_axis in rec:
        data[rec[x_axis]] = data.get(x_axis, 0) + float(rec.get(y_column, 0))
    self.parent.context.register(map_chart)
    locations = list(data.keys())
    map_chart.add_trace({'locations': locations, 'z': [data[k] for k in locations]})
    map_chart.data.locationmode = 'country names'
    map_chart.layout.geo.scope = 'africa'
    map_chart.layout.geo.showlakes = True
    map_chart.layout.geo.showland = True
    return map_chart

  def usa(self, record, y_column=None, x_axis=None, title=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Choropleth`

    Related Pages:

      https://plotly.com/javascript/choropleth-maps/

    :param record:
    :param y_column:
    :param x_axis:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    map_chart = geo.GeoPlotly.Choropleth(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    map_chart.options.responsive = True
    data = {}
    for rec in record:
      if x_axis in rec:
        data[rec[x_axis]] = data.get(x_axis, 0) + float(rec.get(y_column, 0))
    self.parent.context.register(map_chart)
    locations = list(data.keys())
    map_chart.add_trace({'locations': locations, 'z': [data[k] for k in locations]})
    map_chart.data.locationmode = 'USA-states'
    map_chart.layout.geo.scope = 'usa'
    map_chart.layout.geo.showlakes = True
    map_chart.layout.geo.showland = True
    return map_chart




from epyk.core.html.graph import GraphPlotly

from epyk.core.js.packages import JsPlotly


class DataMarkersChoroMapBar(GraphPlotly.DataMarkers):

  @property
  def title(self):
    return self._attrs["title"]

  @title.setter
  def title(self, val):
    self._attrs["title"] = val

  @property
  def ticksuffix(self):
    return self._attrs["ticksuffix"]

  @ticksuffix.setter
  def ticksuffix(self, val):
    self._attrs["ticksuffix"] = val

  @property
  def showticksuffix(self):
    return self._attrs["showticksuffix"]

  @showticksuffix.setter
  def showticksuffix(self, val):
    self._attrs["showticksuffix"] = val


class DataMarkersChoroMap(GraphPlotly.DataMarkers):

  @property
  def cmax(self):
    return self._attrs["cmax"]

  @cmax.setter
  def cmax(self, val):
    self._attrs["cmax"] = val

  @property
  def cmin(self):
    return self._attrs["cmin"]

  @cmin.setter
  def cmin(self, val):
    self._attrs["cmin"] = val

  @property
  def colorscale(self):
    return self._attrs["colorscale"]

  @colorscale.setter
  def colorscale(self, val):
    self._attrs["colorscale"] = val

  @property
  def colorbar(self):
    """

    :rtype: LayoutMapBox
    """
    return self.sub_data("colorbar", DataMarkersChoroMapBar)


class LayoutCenter(GraphPlotly.Layout):

  @property
  def lat(self):
    return self._attrs["lat"]

  @lat.setter
  def lat(self, val):
    self._attrs["lat"] = val

  @property
  def lon(self):
    return self._attrs["lon"]

  @lon.setter
  def lon(self, val):
    self._attrs["lon"] = val


class LayoutMapBox(GraphPlotly.Layout):

  @property
  def style(self):
    return self._attrs["style"]

  @style.setter
  def style(self, val):
    self._attrs["style"] = val

  @property
  def zoom(self):
    return self._attrs["zoom"]

  @zoom.setter
  def zoom(self, val):
    self._attrs["zoom"] = val

  @property
  def center(self):
    """

    :rtype: LayoutCenter
    """
    return self.sub_data("center", LayoutCenter)


class LayoutGeo(GraphPlotly.Layout):

  @property
  def dragmode(self):
    return self._attrs["dragmode"]

  @dragmode.setter
  def dragmode(self, val):
    self._attrs["dragmode"] = val

  @property
  def mapboxAccessToken(self):
    return self._attrs["mapboxAccessToken"]

  @mapboxAccessToken.setter
  def mapboxAccessToken(self, val):
    self._attrs["mapboxAccessToken"] = val

  @property
  def mapbox(self):
    """

    :rtype: LayoutMapBox
    """
    return self.sub_data("mapbox", LayoutMapBox)


class LayoutGeoMapGeo(LayoutGeo):

  @property
  def scope(self):
    return self._attrs["scope"]

  @scope.setter
  def scope(self, val):
    self._attrs["scope"] = val

  @property
  def resolution(self):
    return self._attrs["resolution"]

  @resolution.setter
  def resolution(self, val):
    self._attrs["resolution"] = val

  @property
  def showlakes(self):
    return self._attrs["showlakes"]

  @showlakes.setter
  def showlakes(self, val):
    self._attrs["showlakes"] = val

  @property
  def oceancolor(self):
    return self._attrs["oceancolor"]

  @oceancolor.setter
  def oceancolor(self, val):
    self._attrs["oceancolor"] = val

  @property
  def showland(self):
    return self._attrs["showland"]

  @showland.setter
  def showland(self, val):
    self._attrs["showland"] = val

  @property
  def showframe(self):
    return self._attrs["showframe"]

  @showframe.setter
  def showframe(self, val):
    self._attrs["showframe"] = val

  @property
  def bgcolor(self):
    return self._attrs["bgcolor"]

  @bgcolor.setter
  def bgcolor(self, val):
    self._attrs["bgcolor"] = val

  @property
  def lakecolor(self):
    return self._attrs["lakecolor"]

  @lakecolor.setter
  def lakecolor(self, val):
    self._attrs["lakecolor"] = val

  @property
  def landcolor(self):
    return self._attrs["landcolor"]

  @landcolor.setter
  def landcolor(self, val):
    self._attrs["landcolor"] = val


class LayoutGeoMap(LayoutGeo):

  @property
  def geo(self):
    """

    :rtype: LayoutGeoMapGeo
    """
    return self.sub_data("geo", LayoutGeoMapGeo)

  @property
  def mapbox(self):
    """

    :rtype: LayoutMapBox
    """
    return self.sub_data("mapbox", LayoutMapBox)

  def no_background(self):
    """

    https://community.plot.ly/t/you-can-remove-the-white-background-of-the-graphics-background/933

    :return:
    """
    self.paper_bgcolor = 'rgba(0,0,0,0)'
    self.plot_bgcolor = 'rgba(0,0,0,0)'
    self.geo.bgcolor = 'rgba(0,0,0,0)'
    return self

class DataScatterMapBox(GraphPlotly.DataChart):

  @property
  def lon(self):
    return self._attrs["lon"]

  @lon.setter
  def lon(self, val):
    self._attrs["lon"] = val

  @property
  def lat(self):
    return self._attrs["lat"]

  @lat.setter
  def lat(self, val):
    self._attrs["lat"] = val


class DataChoropleth(GraphPlotly.DataChart):
  @property
  def locationmode(self):
    return self._attrs["locationmode"]

  @locationmode.setter
  def locationmode(self, val):
    self._attrs["locationmode"] = val

  @property
  def autocolorscale(self):
    return self._attrs["autocolorscale"]

  @autocolorscale.setter
  def autocolorscale(self, val):
    self._attrs["autocolorscale"] = val

  @property
  def zmin(self):
    return self._attrs["zmin"]

  @zmin.setter
  def zmin(self, val):
    self._attrs["zmin"] = val

  @property
  def zmax(self):
    return self._attrs["zmax"]

  @zmax.setter
  def zmax(self, val):
    self._attrs["zmax"] = val

  @property
  def marker(self):
    """

    https://plot.ly/javascript/bubble-maps/
    """
    return self.sub_data("marker", DataMarkersChoroMap)


class DataBubble(DataChoropleth):

  @property
  def locations(self):
    return self._attrs["locations"]

  @locations.setter
  def locations(self, val):
    self._attrs["locations"] = val


class Scatter(GraphPlotly.Chart):

  __reqJs = ['plotly.js']

  @property
  def chart(self):
    """
    :rtype: JsPlotly.Bar
    """
    if self._chart is None:
      self._chart = JsPlotly.Pie(self._report, varName=self.chartId)
    return self._chart

  @property
  def layout(self):
    """

    :rtype: LayoutGeo
    """
    if self._layout is None:
      self._layout = LayoutGeo(self._report)
    return self._layout

  @property
  def data(self):
    return self._traces[-1]

  def add_trace(self, data, type='scattermapbox', mode=None):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataScatterMapBox(self._report, attrs=c_data))
    return self


class Chorolet(GraphPlotly.Chart):

  __reqJs = ['plotly.js']

  @property
  def chart(self):
    """
    :rtype: JsPlotly.Bar
    """
    if self._chart is None:
      self._chart = JsPlotly.Pie(self._report, varName=self.chartId)
    return self._chart

  @property
  def layout(self):
    """

    :rtype: LayoutGeo
    """
    if self._layout is None:
      self._layout = LayoutGeo(self._report)
    return self._layout

  @property
  def data(self):
    return self._traces[-1]

  def add_trace(self, data, type='choroplethmapbox', mode=None):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataScatterMapBox(self._report, attrs=c_data))
    return self


class Choropleth(GraphPlotly.Chart):

  __reqJs = ['plotly.js']

  @property
  def chart(self):
    """
    :rtype: JsPlotly.Bar
    """
    if self._chart is None:
      self._chart = JsPlotly.Pie(self._report, varName=self.chartId)
    return self._chart

  @property
  def layout(self):
    """

    :rtype: LayoutGeo
    """
    if self._layout is None:
      self._layout = LayoutGeoMap(self._report)
    return self._layout

  @property
  def data(self):
    return self._traces[-1]

  def add_trace(self, data, type='choropleth', mode=None):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataChoropleth(self._report, attrs=c_data))
    return self


class ScatterGeo(GraphPlotly.Chart):

  __reqJs = ['plotly.js']

  @property
  def chart(self):
    """
    :rtype: JsPlotly.Bar
    """
    if self._chart is None:
      self._chart = JsPlotly.Pie(self._report, varName=self.chartId)
    return self._chart

  @property
  def layout(self):
    """

    :rtype: LayoutGeo
    """
    if self._layout is None:
      self._layout = LayoutGeoMap(self._report)
    return self._layout

  @property
  def data(self):
    return self._traces[-1]

  def add_trace(self, data, type='scattergeo', mode='markers'):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataChoropleth(self._report, attrs=c_data))
    return self


class BubbleGeo(GraphPlotly.Chart):

  __reqJs = ['plotly.js']

  @property
  def chart(self):
    """
    :rtype: JsPlotly.Bar
    """
    if self._chart is None:
      self._chart = JsPlotly.Pie(self._report, varName=self.chartId)
    return self._chart

  @property
  def layout(self):
    """

    :rtype: LayoutGeo
    """
    if self._layout is None:
      self._layout = LayoutGeoMap(self._report)
    return self._layout

  @property
  def data(self):
    return self._traces[-1]

  def add_trace(self, data, type='scattergeo', mode='markers'):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataBubble(self._report, attrs=c_data))
    return self



import os
import time
import json

from epyk.core.js import Imports
from epyk.core.js import Js
from epyk.core.js import JsUtils
from epyk.core.js.Imports import requires

from epyk.core.html.templates import HtmlTmplBase


class OutBrowsers(object):
  def __init__(self, context):
    self._context = context

  def codepen(self, path=None, target="_blank", open_browser=True):
    """
    Description:
    ------------
    Update the Html launcher and send the data to codepen
    URL used: https://codepen.io/pen/define/

    Related Pages:

			https://www.debuggex.com/cheatsheet/regex/python

    Attributes:
    ----------
    :param path: Output path in which the static files will be generated
    :param target: String. Load the data in a new tab in the browser
    :param open_browser: Boolean. Flag to open the browser automatically

    :return: The output launcher full file name
    """
    import re
    import webbrowser

    results = self._context._to_html_obj()
    js_external = re.findall('<script language="javascript" type="text/javascript" src="(.*?)"></script>', results['jsImports'])
    css_external = re.findall('<link rel="stylesheet" href="(.*?)" type="text/css">', results['cssImports'])
    jsObj = Js.JsBase()
    result = {"js": results["jsFrgs"], "js_external": ";".join(js_external), "css_external": ";".join(css_external), "html": results['content'], "css": results["cssStyle"]}
    data = jsObj.location.postTo("https://codepen.io/pen/define/", {"data": json.dumps(result)}, target=target)
    if path is None:
      path = os.path.join(os.getcwd(), "outs")
    else:
      path = os.path.join(path)
    if not os.path.exists(path):
      os.makedirs(path, exist_ok=True)
    with open(os.path.join(path, "RunnerCodepen.html"), "w") as f:
      f.write('<html><body></body><script>%s</script></html>' % data.replace("\\\\n", ""))
    launcher_file = os.path.join(path, "RunnerCodepen.html")
    if open_browser:
      webbrowser.open(launcher_file)
    return launcher_file


class PyOuts(object):
  def __init__(self, report=None, options=None):
    self._report, self._options = report, options
    self.excluded_packages = None

  def _to_html_obj(self, htmlParts=None, cssParts=None):
    """
    Description:
    ------------
    Create the HTML result object from the report definition

    Attributes:
    ----------
    :param htmlParts: Optional. HTML Content of the page
    :param cssParts: Optional. CSS classes content of the page

    :return: A python dictionary with the HTML results
    """
    if htmlParts is None:
      htmlParts, cssParts = [], {}
      for objId in self._report.content:
        #if content_only and self._report.htmlItems[objId].name == "Nav Bar":
        #  continue

        if self._report.htmlItems[objId].inReport:
          htmlParts.append(self._report.htmlItems[objId].html())
        #
        cssParts.update(self._report.htmlItems[objId].style.get_classes_css())
    onloadParts = list(self._report._jsText)
    for data_id, data in self._report._props.get("data", {}).get('sources', {}).items():
      onloadParts.append("var data_%s = %s" % (data_id, json.dumps(data)))

    for k, v in self._report._props.get('js', {}).get('functions', {}).items():
      sPmt = "(%s)" % ", ".join(list(v["pmt"])) if "pmt" in v else "{}"
      onloadParts.append("function %s%s{%s}" % (k, sPmt, v["content"].strip()))

    for c, d in self._report._props.get('js', {}).get("constructors", {}).items():
      onloadParts.append(d)

    for c, d in self._report._props.get('js', {}).get("datasets", {}).items():
      onloadParts.append(d)

    for b in self._report._props.get('js', {}).get("builders", []):
      onloadParts.append(b)

    for b in self._report._props.get('js', {}).get("builders_css", []):
      onloadParts.append(b)

    # Add the component on ready functions
    for objId in self._report.content:
      obj_id = self._report.htmlItems[objId].dom.varId
      if obj_id and obj_id in self._report._props.get('js', {}).get('onCompReady', {}):
        onloadParts.append(self._report._props['js']['onCompReady'][obj_id])
      for event, fncs in self._report.htmlItems[objId]._events['doc_ready'].items():
        if fncs['profile']:
          # Add the profiling feature if the variable is set to true
          profile_var = self._report.js.performance.add_profiling(fncs['content'])
          fncs['content'].append(self._report.js.console.log(self._report.js.objects.get(profile_var)))
        str_fncs = JsUtils.jsConvertFncs(fncs['content'], toStr=True)
        onloadParts.append("%s.addEventListener('%s', function(event){%s})" % (obj_id, event, str_fncs))

    # Add the page on document ready functions
    for on_ready_frg in self._report._props.get('js', {}).get('onReady', []):
      onloadParts.append(on_ready_frg)

    importMng = Imports.ImportManager(online=True, report=self._report)
    results = {
      'cssStyle': "%s\n%s" % ("\n".join([v for v in cssParts.values()]), "\n".join(self._report._cssText)),
      'cssContainer': ";".join(["%s:%s" % (k, v) for k, v in self._report._props.get('css', {}).get('container', {}).items()]),
      'content': "\n".join(htmlParts),
      'jsFrgs': ";".join(onloadParts),
      'cssImports': importMng.cssResolve(self._report.cssImport, self._report.cssLocalImports, excluded=self.excluded_packages),
      'jsImports': importMng.jsResolve(self._report.jsImports, self._report.jsLocalImports, excluded=self.excluded_packages)
    }
    return results

  def _repr_html_(self):
    """
    Description:
    ------------
    Standard output for Jupyter Notebooks.

    This is what will use IPython in order to display the results in cells.
    """
    results = self._to_html_obj()
    importMng = Imports.ImportManager(online=True, report=self._report)
    require_js = importMng.to_requireJs(results, self.excluded_packages)
    results['paths'] = "{%s}" % ", ".join(["%s: '%s'" % (k, p) for k, p in require_js['paths'].items()])
    results['jsFrgs_in_req'] = require_js['jsFrgs']
    return HtmlTmplBase.JUPYTER.strip() % results

  def jupyterlab(self):
    """
    Description:
    ------------
    For a display of the report in JupyterLab.
    Thanks to this function some packages will not be imported to not conflict with the existing ones

    Related Pages:

			https://jupyter.org/
    """
    self.excluded_packages = ['bootstrap']
    return self

  def jupyter(self):
    """
    Description:
    ------------
    For a display of the report in Jupyter.
    Thanks to this function some packages will not be imported to not conflict with the existing ones

    Related Pages:

			https://jupyter.org/

    :return: The ouput object with the function _repr_html_
    """
    self.excluded_packages = ['bootstrap', 'jquery', 'moment']
    return self

  def w3cTryIt(self, path=None, name=None):
    """
    Description:
    ------------
    This will produce everything in a single page which can be directly copied to the try editor in w3C website

    Related Pages:

			https://www.w3schools.com/html/tryit.asp?filename=tryhtml_basic

    Attributes:
    ----------
    :param path: The path in which the output files will be created
    :param name: The filename without the extension

    """
    if path is None:
      path = os.path.join(os.getcwd(), "outs", "w3schools")
    else:
      path = os.path.join(path, "w3schools")
    if not os.path.exists(path):
      os.makedirs(path, exist_ok=True)
    if name is None:
      name = int(time.time())
    file_path = os.path.join(path, "%s.html" % name)
    with open(file_path, "w") as f:
      f.write(self._repr_html_())
    return file_path

  def codepen(self, path=None, name=None):
    """
    Description:
    ------------

    Usage::

      Related Pages:

			https://codepen.io/

    Attributes:
    ----------
    :param path: The path in which the output files will be created
    :param name: The filename without the extension

    TODO Try to add the prefill
    https://blog.codepen.io/documentation/api/prefill/

    :return: The file path
    """
    self.jsfiddle(path, name, framework="codepen")

  def jsfiddle(self, path=None, name=None, framework="jsfiddle"):
    """
    Description:
    ------------
    Produce files which can be copied directly to https://jsfiddle.net in order to test the results and perform changes.

    The output is always in a sub directory jsfiddle

    Usage::

      Related Pages:

			https://jsfiddle.net/

    Attributes:
    ----------
    :param path: The path in which the output files will be created
    :param name: The filename without the extension

    :return: The file path
    """
    if path is None:
      path = os.path.join(os.getcwd(), "outs", framework)
    else:
      path = os.path.join(path, framework)
    if not os.path.exists(path):
      os.makedirs(path, exist_ok=True)
    if os.path.exists(path):
      if name is None:
        name = int(time.time())
      results = self._to_html_obj(content_only=True)
      # For the JavaScript builders
      with open(os.path.join(path, "%s.js" % name), "w") as f:
        f.write(results["jsFrgs"])

      # FOr all the doms and imports
      with open(os.path.join(path, "%s.html" % name), "w") as f:
        f.write("%s\n" % results["cssImports"])
        f.write("%s\n" % results["jsImports"])
        f.write(results["content"])

      # For the CSS styles
      with open(os.path.join(path, "%s.css" % name), "w") as f:
        f.write(results["cssStyle"])
    return path

  def html_file(self, path=None, name=None):
    """
    Description:
    ------------
    Function used to generate a static HTML page for the report

    Usage::

      Attributes:
    ----------
    :param path: The path in which the output files will be created
    :param name: The filename without the extension

    :return: The file full path
    """
    if path is None:
      path = os.path.join(os.getcwd(), "outs")
    if not os.path.exists(path):
      os.makedirs(path, exist_ok=True)
    if name is None:
      name = int(time.time())
    file_path = os.path.join(path, "%s.html" % name)
    with open(file_path, "w") as f:
      htmlParts = []
      cssParts = dict(self._report.body.style.get_classes_css())
      for objId in self._report.content:
        if self._report.htmlItems[objId].inReport:
          htmlParts.append(self._report.htmlItems[objId].html())
        cssParts.update(self._report.htmlItems[objId].style.get_classes_css())
      body = str(self._report.body.set_content(self._report, "\n".join(htmlParts)))
      results = self._to_html_obj(htmlParts, cssParts)
      results['body'] = body
      results['header'] = self._report.headers
      f.write(HtmlTmplBase.STATIC_PAGE % results)
    return file_path

  def markdown_file(self, path=None, name=None):
    """
    Description:
    ------------
    Writes a Markdown file from the report object

    Attributes:
    ----------
    :param path: The path in which the output files will be created
    :param name: The filename without the extension

    :return: The file path
    """
    if path is None:
      path = os.path.join(os.getcwd(), "outs", "markdowns")
    else:
      path = os.path.join(path, "markdowns")
    if not os.path.exists(path):
      os.makedirs(path, exist_ok=True)
    if os.path.exists(path):
      if name is None:
        name = "md_%s.amd" % int(time.time())

      file_Path = os.path.join(path, name)
      with open(file_Path, "w") as f:
        for objId in self._report.content:
          html_obj = self._report.htmlItems[objId]
          if hasattr(html_obj, "to_markdown"):
            f.write("%s\n" % html_obj.to_markdown(html_obj.vals))
      return file_Path

  def str(self):
    """

    :return:
    """

  def pdf(self):
    """

    :return:
    """
    self.word()
    # Then save the file to pdf

  def word(self, path=None, name=None):
    """
    Description:
    ------------
    Writes the result to an Word document

    Attributes:
    ----------
    :param path: The path in which the output files will be created
    :param name: The filename without the extension

    :return:
    """
    from docx import Document
    from docx.shared import RGBColor

    if path is None:
      path = os.path.join(os.getcwd(), "outs", "office")
    else:
      path = os.path.join(path, "office")
    if not os.path.exists(path):
      os.makedirs(path, exist_ok=True)
    timestamp = time.strftime("%Y%m%d_%H%M%S", time.gmtime())
    name = name or 'word_%s.docx' % timestamp
    document = Document()
    for objId in self._report.content:
      if self._report.htmlItems[objId].inReport:
        try:
          self._report.htmlItems[objId].to_word(document)
        except Exception as err:
          error_title = document.add_heading().add_run("Error")
          error_title.font.color.rgb = RGBColor(255, 0, 0)
          error_title.font.italic = True
          error_paragraph = document.add_paragraph().add_run((str(err)))
          error_paragraph.font.color.rgb = RGBColor(255, 0, 0)
          error_paragraph.font.italic = True
    document.save(os.path.join(path, name))
    return name

  def excel(self, path=None, name=None):
    """
    Description:
    ------------
    Writes the result to an Excel document

    Attributes:
    ----------
    :param path: The path in which the output files will be created
    :param name: The filename without the extension

    :return:
    """
    xls = requires("xlsxwriter", reason='Missing Package', install='xlsxwriter', source_script=__file__)

    if path is None:
      path = os.path.join(os.getcwd(), "outs", "office")
    else:
      path = os.path.join(path, "office")
    if not os.path.exists(path):
      os.makedirs(path, exist_ok=True)
    timestamp = time.strftime("%Y%m%d_%H%M%S", time.gmtime())
    name = name or '%s_%s.xlsx' % (self._report.run.script_name, timestamp)
    workbook = xls.Workbook(os.path.join(path, name))
    worksheet = workbook.add_worksheet()
    cursor = {'row': 0, 'col': 0}
    for objId in self._report.content:
      if self._report.htmlItems[objId].inReport:
        try:
          self._report.htmlItems[objId].to_xls(workbook, worksheet, cursor)
        except Exception as err:
          cell_format = workbook.add_format({'bold': True, 'font_color': 'red'})
          worksheet.write(cursor['row'], 0, str(err), cell_format)
          cursor['row'] += 2
    workbook.close()
    return name

  def power_point(self):
    """
    Description:
    ------------

    :return:
    """

  @property
  def browser(self):
    """
    Description:
    ------------
    This module will require the package webbrowser.
    It will allow outputs to be created directly in the webpages (without using intermediary text files
    """
    return OutBrowsers(self)





from epyk.core.js.packages import JsPackage
from epyk.core.js.primitives import JsObjects


class Json(JsPackage):
  lib_alias = {"js": "json-formatter", 'css': "json-formatter"}

  def openAtDepth(self, n):
    """
    Default: 1 This number indicates up to how many levels the rendered tree should open.
    It allows use cases such as collapse all levels (with value 0) or expand all levels (with value Infinity).

    Attributes:
    ----------
    :param n: Integer. the depth of the tree at start
    """
    return JsObjects.JsObjects.get('%s.openAtDepth(%s)' % (self.varName, n))



"""

"""

import json
import io
import zipfile

from epyk.core.html import Html

# The list of CSS classes
# from epyk.core.css.styles import CssGrpClsFile
# from epyk.core.css.styles import GrpClsButton


class DownloadMemoryZip(Html.Html):
  """

  TODO Find a way to send the in memory file form a report: data: %(archive)s,
  """
  alias, cssCls, __reqCss = 'anchorFMemory', ['btn', 'btn-success'], ['font-awesome', 'bootstrap']
  name, category = 'Memory Files', 'System'
  builder_name = False
  # CSS Classes
  #_grpCls = CssGrpClsButton.CssClassButton

  def __init__(self, report, vals, fileName, cssCls, cssAttr, profile):
    super(DownloadMemoryZip, self).__init__(report, vals,  cssCls, cssAttr, profile=profile)
    self.fileName = fileName
    self.memory_file = io.BytesIO()
    self.zf = zipfile.ZipFile(self.memory_file, mode='w', compression=zipfile.ZIP_DEFLATED)

  def add(self, data, filename):
    """ Add the content of string to a file in the in-memory package

    :param data: The data
    :param filename: The filename

    :return:
    """
    self.zf.writestr(filename, data)

  def namelist(self):
    """ Return the list of files in the in-memory zip archive

    :return:
    """
    return self.zf.namelist()

  def __str__(self):
    self.click('''
        $.ajax({url: %(url)s, type: "POST", contentType: attr("enctype", "multipart/form-data"), data: %(archive)s, success: success})
        ''' % {'htmlId': self.htmlId, 'url': "''", 'archive': self.zf})
    return '<button %s>%s</button>' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.val)


class DropFile(Html.Html):
  __reqCss, __reqJs = ['bootstrap', 'font-awesome'], ['bootstrap']
  name, category, inputType, callFnc = 'Drop File Area', 'Input', "file", 'dropfile'
  #_grpCls = CssGrpClsFile.CssStylesDrop

  def __init__(self, report, vals, tooltip, report_name, file_type, profile):
    super(DropFile, self).__init__(report, vals, profile=profile)
    self.tooltip(tooltip, location='bottom')
    self.report_name, self.dataType = report_name if report_name is not None else self._report.run.report_name, file_type
    for action in ["dragover", "dragleave", "dragenter"]:
      self.jsFrg(action, "event.originalEvent.preventDefault(); event.originalEvent.stopPropagation(); event.originalEvent.dataTransfer.dropEffect = 'copy';")
    self.css({"display": "inline-block", "width": '100%'})

  @property
  def jsQueryData(self): return {}

  def drop(self, url=None, jsData=None, jsFncs=None, httpCodes=None, isPyData=True, refresh=True, extensions=None):
    data = []
    if url is None:
      url = "%s/upload/OUTPUTS/%s" % (self._report._urlsApp['epyk-transfer'], self.report_name)
    if jsFncs is None:
      jsFncs = [self._report.jsReloadPage()]
    elif not isinstance(jsFncs, list):
      jsFncs = [jsFncs]

    if jsData is not None:
      for rec in jsData:
        if isinstance(rec, tuple):
          if isPyData:
            data.append("data.append('%s', %s)" % (rec[0], json.dumps(rec[1])))
          else:
            data.append("data.append('%s', %s)" % (rec[0], rec[1]))
        else:
          data.append("data.append('%s', %s)" % (rec.htmlCode, rec.val))
    super(DropFile, self).drop('''
      event.originalEvent.preventDefault(); event.originalEvent.stopPropagation();
      var files = event.originalEvent.dataTransfer.files; var data = new FormData();
      $.each(event.originalEvent.dataTransfer.files, function(i, file) { 
        var fileExt = '.' + file.name.split('.').pop();
        if(%(extensions)s == null) {data.append(file.name, file)} else {
          if(%(extensions)s.indexOf(fileExt) >= 0) { data.append(file.name, file)}}});
      %(jsData)s; %(ajax)s''' % {"jsData": ";".join(data), "extensions": json.dumps(extensions),
                                 "ajax": self._report.jsAjax(url, success=";".join(jsFncs) if refresh else '' ) })
    return self

  def __str__(self):
    return '''
      <div %(strAttr)s><b><i class="fas fa-cloud-upload-alt" style="font-size:20px"></i>&nbsp;&nbsp;%(vals)s</b></div>
      <input id="%(htmlId)s_report" style="display:none;" value="%(envs)s"/>
      ''' % {'htmlId': self.htmlId, 'strAttr': self.get_attrs(pyClassNames=self.style.get_classes()), 'vals': self.vals, 'envs': self.report_name}


class DropConfiguration(Html.Html):
  __reqCss, __reqJs = ['bootstrap', 'font-awesome'], ['bootstrap']
  #_grpCls = CssGrpClsFile.CssStylesDrop
  name, category, inputType, callFnc = 'Drop Configuration Area', 'Input', "file", 'config'

  def __init__(self, report, vals, htmlCode, url, tablename):
    super(DropConfiguration, self).__init__(report, vals, code=htmlCode)
    self.tooltip('Drop your files here', location='bottom')
    self.code = htmlCode
    for action in ["dragover", "dragleave", "dragenter"]:
      self.jsFrg(action, "event.originalEvent.preventDefault(); event.originalEvent.stopPropagation(); event.originalEvent.dataTransfer.dropEffect = 'copy';")
    self.css({"display": "inline-block", "width": '100%'})
    super(DropConfiguration, self).drop('''
          event.originalEvent.preventDefault(); event.originalEvent.stopPropagation();
          var files = event.originalEvent.dataTransfer.files; var data = new FormData();
          $.each(event.originalEvent.dataTransfer.files, function(i, file) { 
            var fileExt = '.' + file.name.split('.').pop() ; data.append(file.name, file) ;}); %(ajax)s; ''' % {
                                    "ajax": self._report.jsAjax('%s/%s/%s/%s/%s' % (url, self._report.run.report_name, self._report.run.script_name, self.code, tablename)
                                                                                 ,success=self._report.jsReloadPage())})


  @property
  def jsQueryData(self): return {}

  def __str__(self):
    return '''
      <div %(strAttr)s><b><i class="fas fa-cloud-upload-alt" style="font-size:20px"></i>&nbsp;&nbsp;%(vals)s</b></div>
      <input id="%(htmlId)s_report" style="display:none;" value="%(envs)s"/>
      ''' % {'htmlId': self.htmlId, 'strAttr': self.get_attrs(pyClassNames=self.style.get_classes()), 'vals': self.vals, 'envs': self._report.run.report_name}




from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObject
from epyk.core.js.primitives import JsObjects

from epyk.core.js.packages import JsPackage
from epyk.core.js.packages import DataAttrs


class JsPlotly(JsPackage):

  def newPlot(self, htmlId, data=None, layout=None, config=None):
    """
    Draws a new plot in an <div> element, overwriting any existing plot. To update an existing plot in a <div>, it is much more efficient to use Plotly.react than to overwrite it.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param htmlId: DOM node or string id of a DOM node
    :param data: array of objects, see documentation (defaults to [])
    :param layout: object, see documentation (defaults to {})
    :param config: object, see documentation (defaults to {})
    """
    htmlId = JsUtils.jsConvertData(htmlId, None)
    data = JsUtils.jsConvertData(data or [], None)
    layout = JsUtils.jsConvertData(layout or {}, None)
    config = JsUtils.jsConvertData(config or {}, None)
    return JsObject.JsObject.get("Plotly.newPlot(%s, %s, %s, %s)" % (htmlId, data, layout, config))

  def react(self, htmlId, data=None, layout=None, config=None):
    """
    Plotly.react has the same signature as Plotly.newPlot above, and can be used in its place to create a plot, but when called again on the same <div> will update it far more efficiently than Plotly.newPlot, which would destroy and recreate the plot. Plotly.react is as fast as Plotly.restyle/Plotly.relayout documented below.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param htmlId: DOM node or string id of a DOM node
    :param data: array of objects, see documentation (defaults to [])
    :param layout: object, see documentation (defaults to {})
    :param config: object, see documentation (defaults to {})
    """
    htmlId = JsUtils.jsConvertData(htmlId, None)
    data = JsUtils.jsConvertData(data or [], None)
    layout = JsUtils.jsConvertData(layout or {}, None)
    config = JsUtils.jsConvertData(config or {}, None)
    return JsObject.JsObject.get("Plotly.react(%s, %s, %s, %s)" % (htmlId, data, layout, config))

  def restyle(self, htmlId, update=None, traceIndices=None):
    """
    This function has comparable performance to Plotly.react and is faster than redrawing the whole plot with Plotly.newPlot.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param htmlId: DOM node or string id of a DOM node
    :param update: object, see below for examples (defaults to {})
    :param traceIndices: array of integer indices into existing value of data (optional, default behaviour is to apply to all traces)
    """
    htmlId = JsUtils.jsConvertData(htmlId, None)
    update = JsUtils.jsConvertData(update or {}, None)
    if traceIndices is None:
      return JsObject.JsObject.get("Plotly.restyle(%s, %s)" % (htmlId, update))

    traceIndices = JsUtils.jsConvertData(traceIndices, None)
    return JsObject.JsObject.get("Plotly.restyle(%s, %s, %s)" % (htmlId, update, traceIndices))

  def relayout(self, htmlId, update=None):
    """
    This function has comparable performance to Plotly.react and is faster than redrawing the whole plot with Plotly.newPlot.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param htmlId: DOM node or string id of a DOM node
    :param update: object, see below for examples (defaults to {})
    """
    htmlId = JsUtils.jsConvertData(htmlId, None)
    update = JsUtils.jsConvertData(update or {}, None)
    return JsObject.JsObject.get("Plotly.relayout(%s, %s)" % (htmlId, update))

  def update(self, htmlId, data_update=None, layout_update=None, traceIndices=None):
    """
    This function has comparable performance to Plotly.react and is faster than redrawing the whole plot with Plotly.newPlot.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param htmlId: DOM node or string id of a DOM node
    :param data_update: object, see Plotly.restyle above (defaults to {})
    :param layout_update: object, see Plotly.relayout above (defaults to {})
    :param traceIndices: array of integer indices into existing value of data, see Plotly.restyle above (optional, default behaviour is to apply to all traces)
    """
    htmlId = JsUtils.jsConvertData(htmlId, None)
    data_update = JsUtils.jsConvertData(data_update or {}, None)
    layout_update = JsUtils.jsConvertData(layout_update or {}, None)
    if traceIndices is None:
      return JsObject.JsObject.get("Plotly.update(%s, %s, %s)" % (htmlId, data_update, layout_update))

    traceIndices = JsUtils.jsConvertData(traceIndices, None)
    return JsObject.JsObject.get("Plotly.update(%s, %s, %s, %s)" % (htmlId, data_update, layout_update, traceIndices))

  def validate(self, data, layout):
    """
    Plotly.validate allows users to validate their input data array and layout object.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param data: array of objects
    :param layout: object
    """
    data = JsUtils.jsConvertData(data, None)
    layout = JsUtils.jsConvertData(layout, None)
    return JsObject.JsObject.get("Plotly.validate(%s, %s)" % (data, layout))

  def makeTemplate(self, figure):
    """
    Plotly.makeTemplate copies the style information from a figure. It does this by returning a template object which can be passed to the layout.template attribute of another figure.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param figure: figure or DOM Node where figure is a plot object, with {data, layout} members. If a DOM node is used it must be a div element already containing a plot.
    """
    figure = JsUtils.jsConvertData(figure, None)
    return JsObject.JsObject.get("Plotly.makeTemplate(%s)" % figure)

  def validateTemplate(self, figure, template):
    """
    Plotly.validateTemplate allows users to Test for consistency between the given figure and a template, either already included in the figure or given separately.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param figure: figure or DOM Node where figure is a plot object, with {data, layout} members.
    :param template: the template, with its own {data, layout}, to test. If omitted, we will look for a template already attached as the plot's
    """
    figure = JsUtils.jsConvertData(figure, None)
    template = JsUtils.jsConvertData(template, None)
    return JsObject.JsObject.get("Plotly.validateTemplate(%s, %s)" % (figure, template))

  def addTraces(self, htmlId, traces, position=None):
    """
    This function has comparable performance to Plotly.react and is faster than redrawing the whole plot with Plotly.newPlot.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param htmlId:
    :param traces:
    :param position:
    """
    htmlId = JsUtils.jsConvertData(htmlId, None)
    traces = JsUtils.jsConvertData(traces, None)
    if position is None:
      return JsObject.JsObject.get("Plotly.addTraces(%s, %s)" % (htmlId, traces))

      position = JsUtils.jsConvertData(position, None)
    return JsObject.JsObject.get("Plotly.addTraces(%s, %s, %s)" % (htmlId, traces, position))

  def deleteTraces(self, htmlId, positions):
    """
    This function has comparable performance to Plotly.react and is faster than redrawing the whole plot with Plotly.newPlot.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param htmlId:
    :param positions:
    """
    htmlId = JsUtils.jsConvertData(htmlId, None)
    positions = JsUtils.jsConvertData(positions, None)
    return JsObject.JsObject.get("Plotly.deleteTraces(%s, %s)" % (htmlId, positions))

  def moveTraces(self, htmlId, currentPosition, destPosition=None):
    """
    This function has comparable performance to Plotly.react and is faster than redrawing the whole plot with Plotly.newPlot.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param htmlId:
    :param currentPosition:
    :param destPosition:
    """
    htmlId = JsUtils.jsConvertData(htmlId, None)
    currentPosition = JsUtils.jsConvertData(currentPosition, None)
    if destPosition is None:
      return JsObject.JsObject.get("Plotly.moveTraces(%s, %s)" % (htmlId, currentPosition))

    return JsObject.JsObject.get("Plotly.moveTraces(%s, %s, %s)" % (htmlId, currentPosition, destPosition))

  def extendTraces(self, htmlId, tracesExtension, indexTraces):
    """
    This function has comparable performance to Plotly.react and is faster than redrawing the whole plot with Plotly.newPlot.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param htmlId:
    :param tracesExtension:
    :param indexTraces:
    """
    htmlId = JsUtils.jsConvertData(htmlId, None)
    tracesExtension = JsUtils.jsConvertData(tracesExtension, None)
    indexTraces = JsUtils.jsConvertData(indexTraces, None)
    return JsObject.JsObject.get("Plotly.extendTraces(%s, %s, %s)" % (htmlId, tracesExtension, indexTraces))

  def prependTraces(self, htmlId, tracesNew, indexTraces):
    """
    This function has comparable performance to Plotly.react and is faster than redrawing the whole plot with Plotly.newPlot.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param htmlId:
    :param tracesNew:
    :param indexTraces:
    """
    htmlId = JsUtils.jsConvertData(htmlId, None)
    tracesNew = JsUtils.jsConvertData(tracesNew, None)
    indexTraces = JsUtils.jsConvertData(indexTraces, None)
    return JsObject.JsObject.get("Plotly.prependTraces(%s, %s, %s)" % (htmlId, tracesNew, indexTraces))

  def animate(self, htmlId, frameOrGroupNameOrFrameList, animationAttributes):
    """
    This allows you to add animation frames to a graphDiv. The group or name attribute of a frame can be used by Plotly.animate in place of a frame object (or array of frame objects). See example here.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param htmlId: DOM node or string id of a DOM node
    :param frameOrGroupNameOrFrameList: A frame to be animated or an array of frames to be animated in sequence.
    :param animationAttributes: An object, see documentation for examples.
    """
    htmlId = JsUtils.jsConvertData(htmlId, None)
    frameOrGroupNameOrFrameList = JsUtils.jsConvertData(frameOrGroupNameOrFrameList, None)
    animationAttributes = JsUtils.jsConvertData(animationAttributes, None)
    return JsObject.JsObject.get("Plotly.animate(%s, %s, %s)" % (htmlId, frameOrGroupNameOrFrameList, animationAttributes))

  def purge(self):
    """
    Using purge will clear the div, and remove any Plotly plots that have been placed in it.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot
    """
    return JsObject.JsObject.get("Plotly.purge()")

  def toImage(self, htmlId, format):
    """
    toImage will generate a promise to an image of the plot in data URL format.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param htmlId:
    :param format:
    """
    htmlId = JsUtils.jsConvertData(htmlId, None)
    format = JsUtils.jsConvertData(format, None)
    return JsObjects.JsPromise("Plotly.toImage(%s, %s)" % (htmlId, format))

  def downloadImage(self, htmlId, format):
    """
    downloadImage will trigger a request to download the image of a Plotly plot.

    https://plot.ly/javascript/plotlyjs-function-reference/#plotlynewplot

    :param htmlId:
    :param format:
    """
    htmlId = JsUtils.jsConvertData(htmlId, None)
    format = JsUtils.jsConvertData(format, None)
    return JsObject.JsObject.get("Plotly.downloadImage(%s, %s)" % (htmlId, format))


class JsPlotlyLegend(object):

  def x(self):
    pass

  def y(self):
    pass

  def bgcolor(self):
    pass

  def bordercolor(self):
    pass

  def traceorder(self):
    pass

  def font(self):
    pass

  def yref(self):
    pass

  def showlegend(self):
    pass


class JsPlotlyAxis(object):

  def title(self):
    pass

  def showline(self):
    pass

  def showticklabels(self):
    pass

  def showgrid(self):
    pass

  def zeroline(self):
    pass

  def range(self):
    pass

  def autorange(self):
    pass


class JsPlotlyMargin(object):

  def autoexpand(self):
    pass

  def l(self):
    pass

  def r(self):
    pass

  def t(self):
    pass


class JsPlotlyAnnotation(object):
  pass


class JsPlotlyLayout(object):

  def __init__(self, attrs):
    self.__attrs = {}

  @property
  def width(self):
    """
    Type: number greater than or equal to 10
    Sets the plot's width (in px).

    https://plot.ly/javascript/reference/#width
    """
    return self.__attrs["width"]

  @width.setter
  def width(self, val):
    self.__attrs["width"] = float(val)

  @property
  def height(self):
    return

  def legend(self):
    pass

  def title(self):
    pass

  def paper_bgcolor(self):
    pass

  def plot_bgcolor(self):
    pass

  def showlegend(self):
    pass

  def bargap(self):
    pass

  def barmode(self):
    pass

  def xaxis(self):
    pass

  def toStr(self):
    return {}


class PlotlyFont(object):

  def size(self):
    pass


class PlotlyMarkers(object):

  def arearatio(self):
    pass

  def sizemin(self):
    pass

  def sizemax(self):
    pass

  def blend(self):
    pass

  def border(self):
    pass

  def color(self):
    pass

  def size(self):
    pass

  def line(self):
    pass

  def sizeref(self):
    pass

  def opacity(self):
    pass

  def sizemode(self):
    pass


class JsPlotlyTrace(object):

  def __init__(self, htmlObj, varName=None, isPyData=True, report=None):
    self._src, self.varName, self._report = htmlObj, varName, report
    self._layout = None

  def mode(self):
    pass

  def type(self):
    pass

  def name(self):
    pass

  def text(self):
    pass

  def line(self):
    pass

  def connectgaps(self):
    pass


class Line(JsPlotlyTrace):
  lib_alias = {'js': 'plotly.js'}

  @property
  def varId(self): return ""

  @property
  def layout(self):
    if self._layout is None:
      self._layout = JsPlotlyLayout({})
    return self._layout

  def add_trace(self, x, y):
    return JsObject.JsObject.get("Plotly.addTraces('%s', {y: %s, x: %s})" % (self._src.htmlId, y, x))

  def deleteTraces(self, i):
    return JsObject.JsObject.get("Plotly.deleteTraces('%s', %s)" % (self._src.htmlId, i))


class Bar(JsPlotlyTrace):
  lib_alias = {'js': 'plotly.js'}

  @property
  def layout(self):
    if self._layout is None:
      self._layout = JsPlotlyLayout({})
    return self._layout


class Pie(JsPlotlyTrace):
  lib_alias = {'js': 'plotly.js'}

  @property
  def layout(self):
    if self._layout is None:
      self._layout = JsPlotlyLayout({})
    return self._layout




from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsNumber
from epyk.core.js.primitives import JsArray
from epyk.core.js.primitives import JsObject
from epyk.core.js.packages import JsPackage

# All the predefined variable types
from epyk.core.js.fncs import JsFncs


class D3ScaleLinear(object):
  """

  """
  def __init__(self, range):
    self.range = range

  def range(self):
    pass


# class D3Html(JsPackage):
#   lib_alias = {"js": 'd3'}
#
#   def __init__(self, src, selector, tag, setVar=False):
#     self.src, self._selector, self.setVar, self.tag = src, selector, setVar, tag
#     self.__js_uniq_fncs, self.varName, self.setVar, self.src = {}, None, setVar, src
#     self._js, self._u, self._js_enums = [[]], {}, {} # a list of list of object definition
#
#   def append(self, htmlType, setVar=True):
#     """
#     If the specified type is a string, appends a new element of this type (tag name) as the last child of each selected element, or before the next following sibling in the update selection if this is an enter selection.
#
#     Documentation
#     https://github.com/d3/d3-selection/blob/v1.4.0/README.md#selecting-elements
#
#     :param htmlType:
#     """
#     cnv_htmlType = JsUtils.jsConvertData(htmlType, None)
#     return D3Html(self.src, "%s.append(%s)" % (self.toStr(), cnv_htmlType), htmlType, setVar=setVar)
#
#   def rappend(self, htmlType):
#     return self.append(htmlType, setVar=False)
#
#   def attr(self, key, val):
#     """
#
#     :param key:
#     :param val:
#     """
#     return self.fnc("attr(%s, %s)" % (JsUtils.jsConvertData(key, None), JsUtils.jsConvertData(val, None)))
#
#   def text(self, data):
#     return self.fnc("text(function(column) { return column; })")
#
#   def html(self, data):
#     return self.fnc("html(function(d) { return d.value; })")
#
#   def selectAll(self, d3Type):
#     self.fnc("selectAll(%s)" % JsUtils.jsConvertData(d3Type, None))
#     return D3Select(self.src, selector=self.toStr())


class D3Select(JsPackage):
  lib_alias = {"js": 'd3'}

  def data(self, datasets=None):
    """
    Description:
    -----------
    Binds the specified array of data with the selected elements,
    returning a new selection that represents the update selection: the elements successfully bound to data.

    The data is specified for each group in the selection.
    If the selection has multiple groups (such as d3.selectAll followed by selection.selectAll), then data should typically be specified as a function.

    Related Pages:

			https://github.com/d3/d3-selection/blob/v1.4.0/README.md#selection_data

    Attributes:
    ----------
    :param datasets:
    """
    if datasets is None:
      self.fnc("data()")
    else:
      #datasets = JsUtils.jsConvertData(datasets, None)
      self.fnc("data(%s)" % datasets)
    return self

  def dataRecord(self, columns):
    """
    Description:
    -----------
    Convert a list of dictionaries to an iteractor which will return ordered arrays based on the columns definition

    Related Pages:
http://bl.ocks.org/gka/17ee676dc59aa752b4e6

    Attributes:
    ----------
    :param columns: List. The columns to be retrieve for each row
    """
    return self.fnc('''data(function(row, i) { var cell = []; %s.forEach(function(k) { cell.push(row[k]);
        }); return cell; }) ''' % JsUtils.jsConvertData(columns, None))

  def dataFncRows(self, columns):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param columns:
    """
    self.fnc("data(function(row) {return %s.map(function(c, i) {return {column: c, value: row[i]}; }); })" % columns)
    return self

  def datum(self, datasets=None):
    """
    Description:
    -----------
    Gets or sets the bound data for each selected element.
    Unlike selection.data, this method does not compute a join and does not affect indexes or the enter and exit selections.

    Related Pages:

			https://github.com/d3/d3-selection/blob/v1.4.0/README.md#selection_datum

    :param datasets: If a value is not specified, returns the bound datum for the first (non-null) element in the selection.
    """
    if datasets is None:
      self.fnc("datum()")
    else:
      self.fnc("datum(%s)" % JsUtils.jsConvertData(datasets, None))
    return self

  def filter(self, fnc):
    pass

  def enter(self):
    """
    Description:
    -----------
    Returns the enter selection: placeholder nodes for each datum that had no corresponding DOM element in the selection.
    (The enter selection is empty for selections not returned by selection.data.)

    Related Pages:

			https://github.com/d3/d3-selection/blob/v1.4.0/README.md#selection_enter
    """
    return self.fnc("enter()")

  def exit(self):
    """
    Description:
    -----------
    Returns the exit selection: existing DOM elements in the selection for which no new datum was found.
    (The exit selection is empty for selections not returned by selection.data.)

    Related Pages:

			https://github.com/d3/d3-selection/blob/v1.4.0/README.md#selection_exit
    """
    return self.fnc("exit()")

  def append(self, htmlType, setVar=True):
    """
    Description:
    -----------
    If the specified type is a string, appends a new element of this type (tag name) as the last child of each selected element, or before the next following sibling in the update selection if this is an enter selection.

    Related Pages:

			https://github.com/d3/d3-selection/blob/v1.4.0/README.md#selecting-elements

    Attributes:
    ----------
    :param htmlType:
    """
    return self.fnc("append(%s)" % JsUtils.jsConvertData(htmlType, None))
    #return D3Html(self.src, "%s.append(%s)" % (self.toStr(), JsUtils.jsConvertData(htmlType, None)), htmlType, setVar=setVar)

  def rappend(self, htmlType):
    """
    Description:
    -----------
    Recursive append function without defining the variable on the javascript side

    Attributes:
    ----------
    :param htmlType:
    """
    return self.append(htmlType, setVar=False)

  def insert(self, htmlType, id):
    pass

  def style(self, key, val):
    pass

  def attr(self, key, val):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param key:
    :param val:
    """
    return self.fnc("attr(%s, %s)" % (JsUtils.jsConvertData(key, None), JsUtils.jsConvertData(val, None)))

  def text(self, data):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param data:
    """
    return self.fnc("text(function(column) { return column; })")

  def html(self, jsData=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsData:
    """
    if jsData is None:
      return self.fnc("html(function(d) { return d; })")

    return self.fnc("html(function(d) { return %s; })" % jsData)

  def htmlByKey(self, data):
    return self.fnc("html(function(d) {console.log('rrr'+ d); return d[%s]; })" % JsUtils.jsConvertData(data, None))

  def call(self, fnc_name):
    """

    Attributes:
    ----------
    :param fnc_name:
    """
    return self.fnc("call(%s)" % JsUtils.jsConvertData(fnc_name, None))

  def on(self, eventType, fnc):
    pass

  def remove(self):
    """
    Description:
    -----------
    Removes the selected elements from the document. Returns this selection (the removed elements) which are now detached from the DOM.
    There is not currently a dedicated API to add removed elements back to the document; however, you can pass a function to selection.append or selection.insert to re-add elements.

    Related Pages:

			https://github.com/d3/d3-selection/blob/v1.4.0/README.md#selecting-elements
    """
    return self.fnc("remove()")

  def clone(self, deep=False):
    """
    Description:
    -----------
    Inserts clones of the selected elements immediately following the selected elements and returns a selection of the newly added clones.

    Related Pages:

			https://github.com/d3/d3-selection/blob/v1.4.0/README.md#selecting-elements

    Attributes:
    ----------
    :param deep: If deep is truthy, the descendant nodes of the selected elements will be cloned as well
    """
    return self.fnc("clone(%s)" % JsUtils.jsConvertData(deep, None))

  def order(self):
    """
    Description:
    -----------
    Re-inserts elements into the document such that the document order of each group matches the selection order.
    This is equivalent to calling selection.sort if the data is already sorted, but much faster.

    Related Pages:

			https://github.com/d3/d3-selection/blob/v1.4.0/README.md#selection_append
    """
    return self.fnc("order()")
  #
  # def select(self, id):
  #   return 'd3.select("%s")' % id

  def selectAll(self, d3Type):
    """

    :param d3Type:
    """
    return self.fnc("selectAll(%s)" % JsUtils.jsConvertData(d3Type, None))

  # def var(self, varName):
  #   """
  #
  #   :param variable_name:
  #   :return:
  #   """
  #   return D3Select(self.src, selector=varName)


class D3ForceSimulation(object):
  _package = ['d3-force']

  def __init__(self, id=None, d3Type=None):
    """

    Documentation:
      - https://github.com/d3/d3-force

    :param id:
    :param d3Type:
    """
    pass

  def force(self, forceType, fnc):
    pass

  def restart(self):
    pass

  def stop(self):
    pass

  def tick(self, tick):
    pass

  def nodes(self, nodes):
    pass

  def alpha(self, number):
    pass

  def alphaDecay(self, number):
    pass

  def alphaTarget(self, target):
    pass

  def alphaTarget(self, number):
    pass

  def velocityDecay(self, decay):
    pass

  def find(self):
    pass

  def on(self):
    pass


class D3ForceManyBody(object):
  def strength(self):
    pass

  def distanceMax(self, x):
    pass

  def distanceMin(self, x):
    pass


class D3ForceCollide(object):
  def __init__(self, radius=None):
    """
    Creates a new circle collision force with the specified radius. If radius is not specified, it defaults to the constant one for all nodes.

    Documentation:
        - https://github.com/d3/d3-force#forceCollide

    :param radius:
    :return:
    """
    pass

  def radius(self, fnc):
    """
    If radius is specified, sets the radius accessor to the specified number or function, re-evaluates the radius accessor for each node, and returns this force. If radius is not specified, returns the current radius accessor

    Documentation:
      - https://github.com/d3/d3-force#forceCollide

    :param fnc:
    :return:
    """
    pass

  def strength(self, strength):
    """
    If strength is specified, sets the force strength to the specified number in the range [0,1] and returns this force. If strength is not specified, returns the current strength which defaults to 0.

    Documentation:
      - https://github.com/d3/d3-force#forceCollide

    :param strength:
    :return:
    """
    pass

  def iterations(self, iterations):
    """
    If iterations is specified, sets the number of iterations per application to the specified number and returns this force

    Documentation:
      - https://github.com/d3/d3-force#forceCollide

    :param iterations:
    :return:
    """
    pass


class D3ForceCenter(object):
  def __init__(self):
    pass

  def x(self):
    pass

  def y(self):
    pass


class D3Event(object):
  def __init__(self):
    """

    Documentation:
      - https://github.com/d3/d3-selection/blob/v1.4.0/README.md#event

    """
    pass

  def on(self):
    pass

  @property
  def active(self):
    pass

  @property
  def x(self):
    pass

  @property
  def y(self):
    pass

  def timeStamp(self):
    pass


class D3Pack(object):
  def __init__(self):
    """

    Documentation:
      - https://github.com/d3/d3-hierarchy/blob/v1.1.8/README.md#pack

    """

  def radius(self, number):
    pass

  def size(self, number):
    """
    If size is specified, sets this pack layout’s size to the specified two-element array of numbers [width, height] and returns this pack layout. If size is not specified, returns the current size, which defaults to [1, 1]

    Documentation:
      - https://github.com/d3/d3-hierarchy/blob/v1.1.8/README.md#pack

    :param number:
    :return:
    """
    pass

  def padding(self, number):
    """
    If padding is specified, sets this pack layout’s padding accessor to the specified number or function and returns this pack layout. If padding is not specified, returns the current padding accessor, which defaults to the constant zero

    Documentation:
      - https://github.com/d3/d3-hierarchy/blob/v1.1.8/README.md#pack

    :param number:
    :return:
    """
    pass

  def packSiblings(self, circles):
    """
    Packs the specified array of circles, each of which must have a circle.r property specifying the circle’s radius

    Documentation:
      - https://github.com/d3/d3-hierarchy/blob/v1.1.8/README.md#pack

    :param circles:
    :return:
    """
    pass

  def packEnclose(self, circles):
    """
    Computes the smallest circle that encloses the specified array of circles, each of which must have a circle.r property specifying the circle’s radius, and circle.x and circle.y properties specifying the circle’s center

    Documentation:
      - https://github.com/d3/d3-hierarchy/blob/v1.1.8/README.md#pack

    :param circles:
    :return:
    """
    pass


class D3Band(object):
  """
  https://github.com/d3/d3-scale/blob/master/README.md#band-scales
  """

  def __init__(self, selector):
    """

    """
    self._selector = selector
    self._js = []

  def domain(self, domain=None):
    """

    :param domain:
    :return:
    """
    if domain is None:
      self._js.append("domain()")
    return self

  def range(self, range=None):
    """

    :param range:
    :return:
    """

  def rangeRound(self, range):
    """

    :param range:
    :return:
    """

  def round(self):
    """

    :return:
    """

  def toStr(self):
    """
    Javascript representation

    :return: Return the Javascript String
    """
    if self._selector is None:
      raise Exception("Selector not defined, use this() or new() first")

    if len(self._js) == 0:
      return self._selector

    strData = "%(jqId)s.%(items)s" % {'jqId': self._selector, 'items': ".".join(self._js)}
    self._js = [] # empty the stack
    return strData


class D3File(object):

  def __init__(self, src, filename, selector):
    self._f, self.src, self._selector = filename, src, selector
    self._js_frg, self._js_ids = [], set()

  def records(self, jsId):
    """

    :param jsId:
    """
    self._js_frg.append("%s.push(row)" % jsId)
    return self

  def unpack(self, jsId, column=None):
    """

    :param jsId:
    :param column:
    """
    if not jsId in self._js_ids:
      column = JsUtils.jsConvertData(column, None)
      self._js_frg.append("%s.push(row[%s])" % (jsId, column))
      self._js_ids.add(jsId)
    return JsObject.JsObject(jsId, isPyData=False)

  def filter(self, rules):
    """

    csv.filter("row['direction'] != 'Decreasing'")

    :param rules:
    """
    if not isinstance(rules, list):
      rules = [rules]
    self._js_frg.append("if(%s){ return; }" % "&&".join(rules))
    return self

  def filterCol(self, column, value, type="=="):
    """

    :param value:
    :return:
    """
    column = JsUtils.jsConvertData(column, None)
    value = JsUtils.jsConvertData(value, None)
    self._js_frg.append("if(row[%s] %s %s){ return; }" % (column, type, value))
    return self

  def callback(self, jsFnc):
    """

    :param jsFnc:
    """
    self._js_frg.append(jsFnc)
    return self

  def toStr(self):
    file = JsUtils.jsConvertData(self._f, None)
    return "%s; %s(%s, function(row, err){%s})" % (";".join(["var %s = []" % i for i in self._js_ids]), self._selector, file, ";".join(self._js_frg))


class D3Svg(object):
  def __init__(self, src, selector, varName=None):
    self.src, self._selector, self.varName = src, selector, varName
    self._js = []

  def line(self):
    self._js.append("line()")
    return self

  def x(self):
    self._js.append("x(function(d) { return x(d.date1); })")
    return self

  def y(self):
    self._js.append("y(function(d) { return x(d.date1); })")
    return self

  def toStr(self):
    content = [self._selector] + self._js
    if self.varName is not None:
      return "var %s = %s" % (self.varName, ".".join(content))

    return ".".join(content)


class JsD3(JsPackage):

  lib_alias = {"js": 'd3'}
  lib_selector = 'd3'

  @property
  def svg(self):
    return D3Svg(self.src, selector="%s.svg" % self._selector)

  def csv(self, url):
    """

    """
    #url = JsUtils.jsConvertData(url, None)
    #return JsFncs.JsFunction("d3.csv(%s)" % (url, JsUtils.jsConvertFncs(callback, toStr=True)))
    return D3File(self.src, url, selector="%s.csv" % self._selector)

  def tsv(self, url):
    """

    """
    #url = JsUtils.jsConvertData(url, None)
    #return JsFncs.JsFunction("d3.csv(%s)" % (url, JsUtils.jsConvertFncs(callback, toStr=True)))
    return D3File(self.src, url, selector="%s.tsv" % self._selector)

  def json(self, url):
    """

    """
    #url = JsUtils.jsConvertData(url, None)
    #return JsFncs.JsFunction("d3.csv(%s)" % (url, JsUtils.jsConvertFncs(callback, toStr=True)))
    return D3File(self.src, url, selector="%s.json" % self._selector)

  def dsv(self, url, delimiter):
    """

    """
    #url = JsUtils.jsConvertData(url, None)
    #return JsFncs.JsFunction("d3.csv(%s)" % (url, JsUtils.jsConvertFncs(callback, toStr=True)))
    return D3File(self.src, url, selector="%s.tsv" % self._selector)

  def min(self, dataset, jsFnc):
    return JsNumber.JsNumber("d3.min(%s, %s)" % (dataset, jsFnc))

  def max(self, dataset, jsFnc):
    return JsNumber.JsNumber("d3.max(%s, %s)" % (dataset, jsFnc))

  def select(self, id, varName=None):
    if varName is not None:
      return D3Select(self.src, selector="d3.select('%s')" % id, varName=varName, setVar=True)

    return D3Select(self.src, selector="d3.select('%s')" % id, setVar=False)

  def selectAll(self, d3Type, setVar=False):
    return D3Select(d3Type=d3Type, setVar=setVar)

  def scaleLinear(self, range):
    return D3ScaleLinear(range)

  def scaleBand(self, range=None):
    """
    Constructs a new band scale with the specified domain and range, no padding, no rounding and center alignment.
    If domain is not specified, it defaults to the empty domain. If range is not specified, it defaults to the unit range [0, 1].

    :param range:

    :rtype: D3Band
    :return:
    """
    if range is None:
      self._js.append("scaleBand()")
    return D3Band(self.toStr())

  def forceSimulation(self, nodes=None):
    return D3ForceSimulation(nodes)

  def forceManyBody(self):
    return D3ForceManyBody()

  def scaleOrdinal(self, colors):
    pass

  def forceX(self):
    pass

  def forceY(self):
    pass

  def drag(self):
    pass

  def toStr(self):
    """
    Javascript representation

    :return: Return the Javascript String
    """
    if self._selector is None:
      raise Exception("Selector not defined, use this() or new() first")

    if len(self._js) == 0:
      return self._selector

    strData = "%(jqId)s.%(items)s" % {'jqId': self._selector, 'items': ".".join(self._js)}
    self._js = [] # empty the stack
    return strData




from epyk.core.js import JsUtils
from epyk.core.js.html import JsHtml
from epyk.core.js.primitives import JsObjects


class JsHtmlSwitch(JsHtml.JsHtmlRich):

  @property
  def content(self):
    return JsHtml.ContentFormatters(self._report, "%s.checked" % self._src.checkbox.dom.varName)

  def set_text(self, value, is_on_val=True):
    """
    Description:
    ------------
    Change the value of the text component

    Usage::

      sw = rptObj.ui.buttons.switch()
    rptObj.ui.button("test").click([
      sw.dom.set_text("ok")])

    Attributes:
    ----------
    :param value: String. The new value
    :param is_on_val: Boolean. Change either the on or the off value displayed
    """
    value = JsUtils.jsConvertData(value, None)
    return JsObjects.JsObjects.get(''' if(%(text)s == %(htmlId)s_data.%(switch_type)s){ %(htmlObj)s };
       %(htmlId)s_data.%(switch_type)s = %(value)s
       ''' % {'htmlId': self._src.htmlId, 'switch_type': 'on' if is_on_val else 'off', 'value': value,
              'text': self._src.switch_text.dom.content.toStr(), 'htmlObj': self._src.switch_text.build(value)})


class Tick(JsHtml.JsHtmlRich):

  @property
  def val(self):
    """
    Description:
    ------------

    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s, label: %s, timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (
        self.htmlId, self.content.toStr(), self._src.span.dom.content.toStr()))

  @property
  def content(self):
    """
    Description:
    ------------
    Get the selected content from the Select component
    """
    # the option variable is coming from the Tick class to get the icon details
    return JsHtml.ContentFormatters(self._report, "%s.classList.contains('%s')" % (self._src.icon.dom.varName, self.options['true'].split(" ")[-1]))


class DomSelect(JsHtml.JsHtmlRich):

  @property
  def val(self):
    """
    Description:
    ------------

    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s, text: %s, options_text: %s, timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (
        self.htmlId, self.content.toStr(), self.text, self.options_text))

  @property
  def content(self):
    """
    Description:
    ------------
    Get the selected content from the Select component

    """
    return JsHtml.ContentFormatters(self._report, "%s.val()" % self.jquery.varId)

  @property
  def text(self):
    """
    Description:
    ------------
    Get the selected content from the Select component

    """
    return JsObjects.JsObjects.get("%s.find('option:selected').text()" % self.jquery.varId)

  @property
  def index(self):
    """
    Description:
    ------------
    Get the selected content from the Select component

    """
    return JsObjects.JsObjects.get("%s.find('option:selected').index()" % self.jquery.varId)

  @property
  def options_text(self):
    """
    Description:
    ------------
    Get the selected content from the Select component

    """
    return JsObjects.JsObjects.get("%s.text()" % self.jquery.varId)




from epyk.core.html import Html
from epyk.core.js.html import JsHtml


class HtmlGeneric(Html.Html):
  category, name = 'Tags', 'tag'

  def __init__(self, report, tag, text, width, height, htmlCode, tooltip, options, profile):
    self.tag = tag
    super(HtmlGeneric, self).__init__(report, text, htmlCode=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    if tooltip:
      self.tooltip(tooltip)
    if options is not None and not options.get('managed', True):
      self.inReport = False

  def __getitem__(self, i):
    if not isinstance(self.val, list):
      return self.val

    return self.val[i]

  def __add__(self, htmlObj):
    """ Add items to a container """
    htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    if not isinstance(self.val, list):
      self._vals = [] if self.val is None else [self.val]
    if htmlObj is not None:
      self.val.append(htmlObj)
    return self

  @property
  def dom(self):
    """
    Description:
    ------------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtml.JsHtmlRich
    """
    if self._dom is None:
      self._dom = JsHtml.JsHtmlRich(self, report=self._report)
    return self._dom

  @property
  def _js__builder__(self):
    return 'htmlObj.innerHTML = data'

  def __str__(self):
    if isinstance(self.val, list):
      str_val = "".join([v.html() if hasattr(v, 'html') else v for v in self.val])
      return '<%s %s>%s</%s>%s' % (self.tag, self.get_attrs(pyClassNames=self.style.get_classes()), str_val, self.tag, self.helper)

    return '<%s %s>%s</%s>%s' % (self.tag, self.get_attrs(pyClassNames=self.style.get_classes()), self.val, self.tag, self.helper)


class HtmlGenericLInk(HtmlGeneric):
  category, name = 'Tags', 'tag'

  def preview(self, url):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param url:

    :return:
    """
    self.on('mouseenter', [
      self._report.js.request_http("test", 'GET', url).send().onSuccess([
        self._report.js.createElement("div", "popup").innerHTML(self._report.js.object("data")).attr('id', 'popup').css({
          'color': 'red', 'display': 'block', 'background': 'white', 'width': '250px', 'padding': '10px'}).position(),
        self._report.body.dom.appendChild(self._report.js.object("popup"))]
      )
    ])

    self.on('mouseleave', [self._report.js.getElementById("popup").remove()])


class HtmlComment(Html.Html):
  category, name = 'Tags', 'comment'

  def __str__(self):
    return '<!--%s-->' % self.val




from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObject


class JsFor(object):

  def __init__(self, jsIterable, iterVar="i", start=0, step=1, context=None):
    """
    The for statement creates a loop that is executed as long as a condition is true.

    The loop will continue to run as long as the condition is true. It will only stop when the condition becomes false.

    Related Pages:

			https//www.w3schools.com/jsref/jsref_for.asp
    https://www.w3schools.com/js/js_performance.asp

    :param jsIterable:
    :param iterVar:
    :param start:
    :param step:
    :param context:
    """
    self._context = context
    if not hasattr(jsIterable, 'toStr'):
      self.__jsObj = JsObject.JsObject(jsIterable, setVar=True)
    else:
      self.__jsObj = jsIterable
    self._js = [
      "var %s = %s" % (self.__jsObj.varName, jsIterable),
      "var l = %s.length" % self.__jsObj.varName,
      "for(var %(iterVar)s = %(start)s; %(iterVar)s < l; i+=%(step)s)" % {"step": step, 'start': start, 'iterVar': iterVar}]
    self.__iterVar = iterVar
    self.__jsFncs = []

  def js(self, jsFncs, isPyData=False, jsFncVal=None):
    """

    :param jsFncs:
    :param isPyData:
    :param jsFncVal:
    :return:
    """
    if jsFncVal is None:
      jsFncVal = "%s[%s]" % (self.__jsObj.varName, self.__iterVar)
    jsIterable = JsUtils.jsConvertFncs(jsFncs, isPyData, jsFncVal)
    self.__jsFncs.extend(jsIterable)
    return self

  def toStr(self):
    """

    :return:
    """
    #if self.__forid is None:
    #  raise Exception("Javascript For loop not correctly defined")

    strData = ";".join(self._js) + "{%s}"
    strData %= ";".join(self.__jsFncs)
    self.__jsFncs, self.__jsObj = [], None  # empty the stack
    return "".join(strData)





import re
import json

from epyk.core.html import Html
from epyk.core.html.options import OptSliders
from epyk.core.html.options import OptList
from epyk.core.html.HtmlList import Li

from epyk.core.html.entities import EntHtml4

from epyk.core.js import JsUtils
from epyk.core.js.html import JsHtmlJqueryUI
from epyk.core.js.html import JsHtmlList
from epyk.core.js.Imports import requires
from epyk.core.js.packages import JsQuery
from epyk.core.js.packages import JsQueryUi

# The list of CSS classes
from epyk.core.css.styles import GrpClsJqueryUI


class ProgressBar(Html.Html):
  __reqCss, __reqJs = ['jqueryui'], ['jqueryui']
  name, category, callFnc = 'Progress Bar', 'Sliders', 'progressbar'

  def __init__(self, report, number, total, width, height, attrs, helper, options, htmlCode, profile):
    options['max'] = total
    super(ProgressBar, self).__init__(report, number, htmlCode=htmlCode, css_attrs={"width": width, "height": height, 'box-sizing': 'border-box'}, profile=profile)
    self.add_helper(helper)
    self._jsStyles = {'css': {"background": self._report.theme.success[1]}}
    self.__options = OptSliders.OptionsProgBar(self, options)

  @property
  def options(self):
    """
    Description:
    -----------
    The progress bar is designed to display the current percent complete for a process.
    The bar is coded to be flexibly sized through CSS and will scale to fit inside its parent container by default.

    Related Pages:

			https://api.jqueryui.com/progressbar

    :rtype: OptSliders.OptionsProgBar
    """
    return self.__options

  @property
  def _js__builder__(self):
    return '''
      options.value = parseFloat(data); jQuery(htmlObj).progressbar(options).find('div').css(options.css);
      %(jqId)s.progressbar(options).find('div').attr("title", ""+ (parseFloat(data) / options.max * 100).toFixed(2) +"%% ("+ parseFloat(data) +" / "+ options.max +")");
      ''' % {"jqId": JsQuery.decorate_var("htmlObj", convert_var=False)}

  @property
  def js(self):
    """
    Description:
    -----------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    Related Pages:

			https://api.jqueryui.com/progressbar

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsQueryUi.ProgressBar
    """
    if self._js is None:
      self._js = JsQueryUi.ProgressBar(self, report=self._report)
    return self._js

  @property
  def dom(self):
    """
    Description:
    ------------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtml.JsHtmlProgressBar
    """
    if self._dom is None:
      self._dom = JsHtmlJqueryUI.JsHtmlProgressBar(self, report=self._report)
    return self._dom

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<div %s></div>%s' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.helper)

  # -----------------------------------------------------------------------------------------
  #                                    MARKDOWN SECTION
  # -----------------------------------------------------------------------------------------
  @staticmethod
  def matchMarkDown(val):
    return re.match("%%%%([0-9]*)%", val)

  @classmethod
  def convertMarkDown(cls, val, regExpResult, report):
    if report is not None:
      getattr(report, 'progressbar')(regExpResult.group(1))
    return ["report.progressbar(%s)" % regExpResult.group(1)]

  @classmethod
  def to_markdown(self, vals):
    return "%%%%" + str(vals) + "%"


class Menu(Html.Html):
  __reqCss, __reqJs = ['jqueryui'], ['jqueryui']
  name, category, callFnc = 'Menu', 'Menus', 'menu'

  def __init__(self, report, records, width, height, attrs, helper, options, htmlCode, profile):
    super(Menu, self).__init__(report, records, css_attrs={"width": width, "height": height}, profile=profile)
    self.add_helper(helper)
    self.__options = OptSliders.OptionsMenu(self, options)
    self.css({"display": 'block', 'position': 'relative'})

  @property
  def options(self):
    """
    Description:
    -----------

    Related Pages:

			https://api.jqueryui.com/menu

    :rtype: OptSliders.OptionsMenu
    """
    return self.__options

  @property
  def _js__builder__(self):
    return '''
          var jqHtmlObj = jQuery(htmlObj); if (options.clearDropDown) {jqHtmlObj.empty()};
          data.forEach(function(rec){
            if (rec.items != undefined) {
              var li = $('<li></li>'); var div = $('<div>'+ rec.value +'</div>').css({"width": '150px'});
              li.append(div); var ul = $('<ul aria-hidden="true"></ul>'); options.clearDropDown = false;
              %(pyCls)s(ul, rec.items, options); li.append(ul); jqHtmlObj.append(li);
            } else {
              var div = $('<div>'+ rec.value +'</div>').css({"width": '150px'}); var li = $('<li></li>');
              li.append(div); jqHtmlObj.append(li)};
          }); jqHtmlObj.menu(options)''' % {"pyCls": self.__class__.__name__}

  @property
  def js(self):
    """
    Description:
    -----------
    Javascript Functions

    Related Pages:

			https://api.jqueryui.com/menu

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsQueryUi.Menu
    """
    if self._js is None:
      self._js = JsQueryUi.Menu(self, report=self._report)
    return self._js

  @property
  def dom(self):
    """
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object
    :rtype: JsHtml.JsHtmlProgressBar
    """
    if self._dom is None:
      self._dom = JsHtmlJqueryUI.JsHtmlProgressBar(self, report=self._report)
    return self._dom

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<ul %s></ul>%s' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.helper)


class Dialog(Html.Html):
  __reqCss, __reqJs = ['jqueryui'], ['jqueryui']
  name, category, callFnc = 'Menu', 'Menus', 'menu'

  def __init__(self, report, text, width, height, attrs, helper, options, htmlCode, profile):
    super(Dialog, self).__init__(report, text, css_attrs={"width": width, "height": height}, profile=profile)
    self.add_helper(helper)
    self.__options = OptSliders.OptionDialog(self, options)

  @property
  def options(self):
    """
    Description:
    -----------
    Open content in an interactive overlay.

    Related Pages:

			https://jqueryui.com/dialog/

    :rtype: OptSliders.OptionDialog
    """
    return self.__options

  @property
  def _js__builder__(self):
    return "jQuery(htmlObj).empty(); jQuery(htmlObj).append('<p>'+ data +'</p>'); jQuery(htmlObj).dialog(options)"

  @property
  def js(self):
    """
    Description:
    -----------
    Open content in an interactive overlay.

    Related Pages:

			https://jqueryui.com/dialog/

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsQueryUi.Dialog
    """
    if self._js is None:
      self._js = JsQueryUi.Dialog(self, report=self._report)
    return self._js

  @property
  def dom(self):
    """
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object
    :rtype: JsHtml.JsHtmlProgressBar
    """
    if self._dom is None:
      self._dom = JsHtmlJqueryUI.JsHtmlProgressBar(self, report=self._report)
    return self._dom

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<div %s>%s</div>' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.helper)


class Slider(Html.Html):
  __reqCss, __reqJs = ['jqueryui'], ['jqueryui']
  name, category, callFnc = 'Slider', 'Sliders', 'slider'

  def __init__(self, report, number, min, max, width, height, attrs, helper, options, htmlCode, profile):
    options.update({'max': max, 'min': min})
    super(Slider, self).__init__(report, number, htmlCode=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    self._jsStyles = {'css': {"background": self._report.theme.success[0]}}
    self.__options = OptSliders.OptionsSlider(self, options)
    self.style.css.padding = "0 10px"
    self.style.css.margin = "15px 0"

  @property
  def options(self):
    """

    :rtype: OptSliders.OptionsSlider
    """
    return self.__options

  @property
  def style(self):
    """
    Description:
    ------------
    Property to the CSS Style of the component

    :rtype: GrpClsJqueryUI.ClassSlider
    """
    if self._styleObj is None:
      self._styleObj = GrpClsJqueryUI.ClassSlider(self)
    return self._styleObj

  @property
  def js(self):
    """
    Description:
    -----------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    Related Pages:

			https://api.jqueryui.com/slider

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsQueryUi.Slider
    """
    if self._js is None:
      self._js = JsQueryUi.Slider(self, report=self._report)
    return self._js

  def change(self, jsFnc, profile=None):
    """
    Description:
    -----------
    Triggered after the user slides a handle, if the value has changed; or if the value is changed programmatically via the value method.

    Related Pages:

			https://api.jqueryui.com/slider/#event-change

    Attributes:
    ----------
    :param jsFnc:
    """
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    self._jsStyles["change"] = "function(event, ui){ %s }" % JsUtils.jsConvertFncs(jsFnc, toStr=True)
    return self

  def start(self, jsFnc, profile=None):
    """
    Description:
    -----------
    Triggered when the user starts sliding.

    Related Pages:

			https://api.jqueryui.com/slider/#event-start

    Attributes:
    ----------
    :param jsFnc:
    """
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    self._jsStyles["start"] = "function(event, ui){ %s }" % JsUtils.jsConvertFncs(jsFnc, toStr=True)
    return self

  def slide(self, jsFnc, profile=None):
    """
    Description:
    -----------
    Triggered when the user starts sliding.

    Related Pages:

			https://api.jqueryui.com/slider/#event-slide

    Attributes:
    ----------
    :param jsFnc:
    """
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    self._jsStyles["slide"] = "function(event, ui){ %s }" % JsUtils.jsConvertFncs(jsFnc, toStr=True)
    return self

  def stop(self, jsFnc, profile=None):
    """
    Description:
    -----------
    Triggered after the user slides a handle.

    Related Pages:

			https://api.jqueryui.com/slider/#event-stop

    Attributes:
    ----------
    :param jsFnc:
    """
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    self._jsStyles["stop"] = "function(event, ui){ %s }" % JsUtils.jsConvertFncs(jsFnc, toStr=True)
    return self

  @property
  def dom(self):
    """
    Description:
    -----------
    The Javascript Dom object

    :rtype: JsHtmlJqueryUI.JsHtmlSlider
    """
    if self._dom is None:
      self._dom = JsHtmlJqueryUI.JsHtmlSlider(self, report=self._report)
    return self._dom

  @property
  def _js__builder__(self):
    return ''' options.value = data; %(jqId)s.slider(options).css(options.css)
      ''' % {"jqId": JsQuery.decorate_var("htmlObj", convert_var=False)}

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '''
      <div %(strAttr)s>
        <div style="width:100%%;height:20px">
          <span style="float:left;display:inline-block">%(min)s</span>
          <span style="float:right;display:inline-block">%(max)s</span>
        </div>
        <div id="%(htmlId)s"></div>
      </div>%(helper)s''' % {"strAttr": self.get_attrs(withId=False), "min": self.options.min, "htmlId": self.htmlId,
                             "max": self.options.max, "helper": self.helper}


class Range(Slider):

  @property
  def _js__builder__(self):
    return '''options.values = [Math.min(...data), Math.max(...data)]; %(jqId)s.slider(options).css(options.css)
        ''' % {"jqId": JsQuery.decorate_var("htmlObj", convert_var=False)}


class SliderDate(Slider):

  def __init__(self, report, number, min, max, width, height, attrs, helper, options, htmlCode, profile):
    super(SliderDate, self).__init__(report, number, min, max, width, height, attrs, helper, options, htmlCode, profile)
    self.options.min = min
    self.options.max = max
    self.options.step = 86400

  @property
  def _js__builder__(self):
    return '''
      const minDt = new Date(options.min).getTime() / 1000;
      const maxDt = new Date(options.max).getTime() / 1000;
      
      options.min = minDt; options.max = maxDt;
      options.value = new Date(data).getTime() / 1000;
      %(jqId)s.slider(options).css(options.css)
      ''' % {"jqId": JsQuery.decorate_var("jQuery(htmlObj)", convert_var=False)}

  @property
  def dom(self):
    """
    Description:
    -----------
    The Javascript Dom object

    :rtype: JsHtmlJqueryUI.JsHtmlSlider
    """
    if self._dom is None:
      self._dom = JsHtmlJqueryUI.JsHtmlSliderDate(self, report=self._report)
    return self._dom


class SliderDates(SliderDate):

  @property
  def _js__builder__(self):
    return ''' console.log(options);
      const minDt = new Date(options.min).getTime() / 1000;
      const maxDt = new Date(options.max).getTime() / 1000;

      options.min = minDt; options.max = maxDt;
      options.values = [new Date(data[0]).getTime() / 1000, new Date(data[1]).getTime() / 1000];
      %(jqId)s.slider(options).css(options.css)
      ''' % {"jqId": JsQuery.decorate_var("jQuery(htmlObj)", convert_var=False)}

  @property
  def dom(self):
    """
    Description:
    -----------
    The Javascript Dom object

    :rtype: JsHtmlJqueryUI.JsHtmlSlider
    """
    if self._dom is None:
      self._dom = JsHtmlJqueryUI.JsHtmlSliderDates(self, report=self._report)
    return self._dom


class SkillBar(Html.Html):
  name, category, callFnc = 'Skill Bars', 'Chart', 'skillbars'

  def __init__(self, report, data, y_column, x_axis, title, width, height, htmlCode, colUrl, colTooltip, filters, profile):
    super(SkillBar, self).__init__(report, "", css_attrs={"width": width, "height": height}, htmlCode=htmlCode,
                                   globalFilter=filters, profile=profile)
    self.add_title(title, options={'content_table': False})
    self.innerPyHTML = report.ui.layouts.table(options={"header": False}) # data, y_column, x_axis)
    self.innerPyHTML.inReport = False
    for rec in data:
      value = report.ui.div(EntHtml4.NO_BREAK_SPACE).css({"width": '%spx' % rec[y_column], 'margin-left': "2px",  "background": report.theme.success[0]})
      value.inReport = False
      self.innerPyHTML += [rec[x_axis], value]
      self.innerPyHTML[-1][1].attr["align"] = 'left'
    self.innerPyHTML.style.clear()
    self.css({"margin": '5px 0'})

  @property
  def _js__builder__(self):
    return '''
      htmlObj.empty();
      data.forEach(function(rec, i){
        if (options.colTooltip != undefined) {var tooltip = 'title="'+ rec[options.colTooltip] +'"'} else {var tooltip = ''};
        if (options.colUrl != undefined) {var content = '<a href="'+ rec[options.colUrl] +'" style="color:white">'+ valPerc.toFixed(2) +'%</a>'} 
        else {var content = rec[options.val].toFixed( 2 ) +"%"};
        htmlObj.append('<tr '+ tooltip +'><td style="width:100px;"><p style="margin:2px;text-align:center;word-wrap:break-word;cursor:pointer">'+ rec[options.label] +'</p></td><td style="width:100%"><div style="margin:2px;display:block;height:100%;padding-bottom:5px;padding:2px 0 2px 5px;width:'+ parseInt( rec[options.val] ) +'%;background-color:'+ jsStyles.color +';color:'+ options.fontColor +'">' + content + '</div></td></tr>');
        htmlObj.find('tr').tooltip()})
      '''

  def __str__(self):
    return '<div %s>%s</div>' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.content)

  # -----------------------------------------------------------------------------------------
  #                                    MARKDOWN SECTION
  # -----------------------------------------------------------------------------------------
  @classmethod
  def matchMarkDownBlock(cls, data): return True if data[0].strip() == "---SkillBar" else None

  @staticmethod
  def matchEndBlock(data): return data.endswith("---")

  @classmethod
  def convertMarkDownBlock(cls, data, report):
    """
    :category:
    :rubric:
    :example:
      ---SkillBar
      label|value|color
      Test 1|35|yellow
      Test 2|25|blue
      ---
    :dsc:

    """
    recordSet = []
    header = data[1].strip().split("|")
    for val in data[2:-1]:
      label, value, color = val.strip().split("|")
      recordSet.append({'value': value, 'label': label, 'color': color})
    if report is not None:
      getattr(report, 'skillbars')(recordSet)
    return ["report.skillbars(%s)" % json.dumps(recordSet)]

  @classmethod
  def jsMarkDown(self, vals):
    recordSet = ["---SkillBar", 'label|value|color']
    for rec in vals:
      recordSet.append("%(label)s|%(value)s|%(color)s" % rec)
    recordSet.append("---")
    return "&&".join(recordSet)

  # -----------------------------------------------------------------------------------------
  #                                    EXPORT OPTIONS
  # -----------------------------------------------------------------------------------------
  def to_img(self):
    """

    :return:
    """
    pkg_matplotlib = requires("matplotlib.pyplot", reason='Missing Package', install='matplotlib', source_script=__file__)
    pkg_numpy = requires("numpy", reason='Missing Package', install='numpy', source_script=__file__)

    pkg_matplotlib.rcdefaults()
    fig, ax = pkg_matplotlib.subplots()

    # Example data
    aggDf = self.vals.data.groupby([self.vals.xAxis])[self.vals.seriesNames[0]].sum().reset_index()
    labels = aggDf[self.vals.xAxis]
    values = aggDf[self.vals.seriesNames[0]]

    y_pos = pkg_numpy.arange(len(labels))
    performance = values
    error = pkg_numpy.random.rand(len(labels))
    ax.barh(y_pos, performance, xerr=error, align='center', color=self._report.theme.colors[7], ecolor='black')
    ax.set_yticks(y_pos)
    ax.set_yticklabels(labels)
    ax.invert_yaxis()  # labels read top-to-bottom
    ax.set_xlabel(self.vals.xAxis)
    ax.set_title(self.title)
    return fig

  def to_word(self, document):
    """

    :param document:

    :return:
    """
    pkg_matplotlib = requires("matplotlib.pyplot", reason='Missing Package', install='matplotlib', source_script=__file__)
    import os

    fig = self.to_img()
    imgsPath = os.path.join(self._report.run.local_path, "imgs")
    if not os.path.exists(imgsPath):
      os.mkdir(imgsPath)

    # Add the picture to the document
    pkg_matplotlib.savefig(os.path.join(imgsPath, "%s.png" % id(fig)))
    document.add_picture(os.path.join(imgsPath, "%s.png" % id(fig)))
    os.remove(os.path.join(imgsPath, "%s.png" % id(fig)))


class ContextMenu(Html.Html):
  name, category, callFnc = 'Context Menu', None, 'contextmenu'
  source = None # The container

  def __init__(self, report, recordSet, width, height, visible, options, profile):
    super(ContextMenu, self).__init__(report, [], css_attrs={"width": width, "height": height}, profile=profile)
    self.__options = OptList.OptionsLi(self, options)
    self.css({'display': 'block' if visible else 'none', 'position': 'absolute', 'z-index': 200,
              'padding': 0, 'margin': 0, 'background-color': self._report.theme.greys[0],
              'border': '1px solid %s' % self._report.theme.colors[5], 'border-radius': '2px'})
    for rec in recordSet:
      self += rec

  @property
  def options(self):
    """

    :rtype: OptList.OptionsLi
    """
    return self.__options

  def __add__(self, htmlObj):
    """

    :param d:
    """
    if not hasattr(htmlObj, 'inReport'):
      htmlObj = self._report.ui.div(htmlObj)
    htmlObj.inReport = False
    li_obj = Li(self._report, htmlObj) if not isinstance(htmlObj, Li) else htmlObj
    li_obj.css({"padding": "5px", 'cursor': 'pointer'})
    if hasattr(htmlObj, 'inReport'):
      htmlObj.inReport = False
    self.val.append(li_obj)
    return self

  def __getitem__(self, i):
    return self.val[i].val

  def onDocumentLoadFnc(self):
    """ Pure Javascript onDocumentLoad Function """
    self.addGlobalFnc("%s(htmlObj, data, jsStyles)" % self.__class__.__name__, '''
      var jsList = htmlObj.find('ul'); jsList.empty();
      data.forEach(function(rec){ 
        if(rec.icon != undefined) {var li = $('<li>').addClass(jsStyles.liStyles).html("<i class='"+ rec.icon +"' style='margin-right:5px'></i>" + rec.text)}
        else {var li = $('<li>').addClass(jsStyles.liStyles).html(rec.text)};
        li.click(function(event){var data = CONTEXT_MENU_VAL; eval(rec.event)}); jsList.append(li)})
      ''', 'Javascript Object builder')

  def __str__(self):
    #self._report._scroll.add("$('nav[name=context_menus]').hide()")
    # TODO: Add a condition in the init to display the context menu only for some columns or rows when table for example
    if getattr(self, 'source') is None:
      raise Exception("Context Menu should be added to a component with the function contextMenu")

    str_vals = "".join([i.html() for i in self.val]) if self.val is not None else ""
    self.mouse(out_fncs=[self.dom.hide()]) # hide when mouse leave the component
    return '''
      <nav %(attr)s name='context_menus'>
        <ul style='list-style:none;padding:0px;margin:0'>%(val)s</ul>
      </nav>''' % {'attr': self.get_attrs(pyClassNames=self.style.get_classes()), 'val': str_vals}


class OptionsBar(Html.Html):
  name, category, callFnc = 'Options', 'Event', 'optionsbar'
  __reqCss, __reqJs = ['font-awesome'], ['font-awesome']
  # _grpCls = CssGrpClsImage.CssClassIcon

  def __init__(self, report, recordset, width, height, size, color, border_color, options):
    super(OptionsBar, self).__init__(report, recordset, width=width[0], widthUnit=width[1], height=height[0], heightUnit=height[1])
    self.css({'padding': '0', 'display': 'block', 'text-align': 'middle', 'color': color, 'margin-left': '5px',
              'background': self._report.theme.greys[0]})
    self.border_color = border_color
    if options.get("draggable", False):
      self.draggable()
    self.size = size

  def draggable(self, options=None):
    self.css({"border": "1px solid %s" % self.border_color})
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.dom.jquery_ui.draggable(options).toStr())
    #self._report.js.addOnLoad(self.dom.jquery_ui.draggable(options).toStr())
    return self

  def __str__(self):
    cssIcon = self._report.style.cssName('CssIcon')
    icons = []
    for rec in self.val:
      rec.update({'cssIcon': cssIcon, 'size': "%s%s" % (self.size[0], self.size[1])})
      if isinstance(rec.get('jsFnc', []), list):
        rec['jsFnc'] = ";".join(rec.get('jsFnc', []))
      if not 'tooltip' in rec:
        rec['tooltip'] = ''
      icons.append('<i class="%(icon)s %(cssIcon)s" style="font-size:%(size)s" title="%(tooltip)s" onclick="var data={event_val:\'%(icon)s\'};%(jsFnc)s"></i>' % rec)
    return '<div %(attrs)s>%(icons)s</div>' % {'attrs': self.get_attrs(pyClassNames=self.defined), 'icons': "".join(icons)}


class SignIn(Html.Html):
  name, category, callFnc = 'SignIn', 'Event', 'signin'
  __reqCss, __reqJs = ['font-awesome'], ['font-awesome']

  def __init__(self, report, text, size, icon):
    super(SignIn, self).__init__(report, text, css_attrs={"width": size, 'height': size})
    self.size, self.icon = "%s%s" % (size[0]-8, size[1]), icon
    self.css({"text-align": "center", "padding": 0, 'color': self._report.theme.colors[3],
              "margin": 0, "border-radius": "%s%s" % (size[0], size[1]),
              "border": "1px solid %s" % self._report.theme.colors[3], 'cursor': 'pointer'})

  def __str__(self):
    if not hasattr(self._report, 'user') or self._report.user == 'local':
      self.attr["class"].add(self.icon or "fas fa-user-tie")
      self.style.css.font_family = "Font Awesome 5 Free"
      self.style.css.padding = "2px"
      self.style.css.font_size = self.size
      return '<i title="Guest Mode" %(attrs)s></i>' % {'size': self.size, 'attrs': self.get_attrs(pyClassNames=self.style.get_classes())}

    return '''
      <div title="%(user)s" %(attrs)s>
        <p style="font-size:%(size)s;margin-top:-2px">%(letter)s</p>
      </div> ''' % {'size': self.size, 'letter': self._report.user[0].upper(), 'user': self._report.user,
                    'attrs': self.get_attrs(pyClassNames=self.style.get_classes())}


class Filters(Html.Html):
  name, category, callFnc = 'Multi Filter', 'Event', 'multiFilter'

  def __init__(self, report, items, title, width, height, htmlCode, helper, options, profile):
    super(Filters, self).__init__(report, [], css_attrs={"width": width, "height": height}, code=htmlCode, profile=profile)
    self.add_title(title, options={'content_table': False, 'managed': False})
    self._jsStyles = {'items': {'display': 'inline-block', 'padding': '1px 4px',
                                'color': self._report.theme.colors[-1], 'margin': '2px', 'border-radius': '5px', 'background-color': self._report.theme.colors[0]}}
    self.__options = OptList.OptionsTagItems(self, options)
    for item in items:
      self += item

  @property
  def options(self):
    """

    :rtype: OptList.OptionsTagItems
    """
    return self.__options

  def __getitem__(self, i):
    return self.val[i]

  def __add__(self, htmlObj):
    """ Add items to a container """
    if not hasattr(htmlObj, 'inReport'):
      htmlObj = self._report.ui.div(htmlObj)
      icon = self._report.ui.icon("fas fa-times").css({"margin-left": '2px'})
      htmlObj += icon
      icon.click([htmlObj.dom.remove()])
    htmlObj.style.css.display = 'inline'
    htmlObj.css(self.options.item_css)
    htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    self.val.append(htmlObj)
    return self

  @property
  def dom(self):
    """
    Description:
    -----------
    The Javascript Dom object

    :rtype: JsHtmlJqueryUI.JsHtmlSlider
    """
    if self._dom is None:
      self._dom = JsHtmlList.Tags(self, report=self._report)
    return self._dom

  def __str__(self):
    str_data = "".join([s.html() for s in self.val])
    return '''
      <div %(cssAttr)s>
        %(title)s
        <div name="panel">%(data)s</div>
      </div>''' % {'title': self.title, 'cssAttr': self.get_attrs(pyClassNames=self.style.get_classes()), 'data': str_data}




import json
import logging

from epyk.core.html import Html

# The list of CSS classes
from epyk.core.css.styles import GrpClsMenu
from epyk.core.css import Defaults_css


class HtmlNavBar(Html.Html):
  name, category = 'Nav Bar', 'System'

  def __init__(self, report, components, width, height, options, profile):
    super(HtmlNavBar, self).__init__(report, [], css_attrs={"width": width, "height": height}, profile=profile)
    if components is not None:
      if not isinstance(components, list):
        components = [components]
      for c in components:
        self.__add__(c)

  @property
  def style(self):
    """
    Description:
    -----------
    Property to the CSS Style of the component

    :rtype: GrpClsMenu.ClassNav
    """
    if self._styleObj is None:
      self._styleObj = GrpClsMenu.ClassNav(self)
    return self._styleObj

  def move(self):
    """

    """
    super(HtmlNavBar, self).move()
    self.style.css.position = None
    self._report.body.style.css.padding_top = 0

  def __add__(self, htmlObj):
    """ Add items to the footer """
    htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    htmlObj.style.css.display = 'inline'
    if htmlObj.css('height') is None:
      htmlObj.style.css.vertical_align = 'middle'
    if htmlObj.css('width') == '100%':
      htmlObj.style.css.width = None
    self.val.append(htmlObj)
    return self

  def add_text(self, text):
    """

    :param text:
    :return:
    """
    val = self._report.ui.text(text)
    self.__add__(val)
    val.style.css.height = "100%"
    val.style.css.vertical_align = 'middle'
    return val

  def __str__(self):
    str_h = "".join([h.html() for h in self.val])
    return "<div %s>%s</div>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_h)


class HtmlParamsBar(Html.Html):
  __reqCss, __reqJs = ['bootstrap'], ['bootstrap']
  # _grpCls = GrpCls.CssGrpClassBase
  allowedComponents = ['date', 'input', 'radio', 'button', 'select', 'title', 'slider', 'switch',
                       'selectmulti', 'delimiter', 'check', 'checkbox', 'inputRange', 'slider', 'inputInt']
  name, category = 'Parameters Bar', 'System'
  docCategory = 'System'

  def __init__(self, report, vals, top, logFiles, height):
    """ Instanciate a container object """
    self.__components, self.__order, self.logFiles = {}, [], logFiles
    self.jsUpdateFnc, self.inCookies = True, {}
    vals = list(vals)
    for component in vals:
      if component['htmlComponent'] not in self.allowedComponents:
        logging.error('%s is not allowed in the parameters bar' % component['htmlComponent'])
        raise Exception('%s is not allowed in the parameters bar' % component['htmlComponent'])

      fnc = getattr(report, component['htmlComponent'])
      parameters = dict(component)
      if 'inCookies' in parameters and report.withCookies:
        self.inCookies[component['htmlCode']] = parameters['inCookies']
        del parameters['inCookies']

      if parameters['htmlComponent'] in ['input', 'date', 'checkbox', 'switch', 'slider']:
        if "widthUnit" not in parameters:
          parameters['widthUnit'] = "px"
        if parameters['htmlComponent'] == 'input':
          parameters['align'] = None
      del parameters['htmlComponent']

      if 'on_init' in parameters:
        recordSet = parameters['on_init'](report)
        parameters['recordSet'] = recordSet
        del parameters['on_init']

      if not component['htmlComponent'] in ['delimiter']:
        htmlCode = component['htmlCode']
        if htmlCode in self.__order:
          raise Exception('code in the Parameters bar - %s - duplicated' % htmlCode)

      htmlObj = fnc(**parameters)
      if hasattr(htmlObj, 'initVal') and htmlCode in report.http:
        htmlObj.initVal(report.http[htmlCode])
      if component['htmlComponent'] in ['delimiter']:
        htmlObj.htmlCode = "component_%s" % id(self)
      self.__add__(htmlObj)
    self.height, self.top = height, top
    report.marginTop += height
    super(HtmlParamsBar, self).__init__(report, [])
    self.css({'margin': 0, "padding": '4px 0', 'display': 'inline-block'})

  def __add__(self, htmlObj):
    """ Add items to a container """
    htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    htmlObj.css({'display': 'inline-block', 'margin-right': '10px'})
    htmlObj.change(htmlObj.jsToUrl())
    # The component added must have a htmlCode class variable defined
    self.__order.append(htmlObj.htmlCode)
    self.__components[htmlObj.htmlCode] = htmlObj
    return self

  def getHtmlCodes(self):
    return self.__order

  def get(self, htmlCode):
    """ Return the Html component in the parameter bar """
    return self.__components[htmlCode]

  def toCookies(self, htmlCode):
    return self.inCookies.get(htmlCode, True)

  def onDocumentReady(self):
    """ Return the javascript calls to be returned to update the component """
    vals = [self.__components[comp].html() for comp in self.__order]
    self._report.jsOnLoadFnc.insert(0, 'NavBar(%s, %s)' % (self.jqId, json.dumps(vals)))

  def onDocumentLoadFnc(self):
    """ Pure Javascript onDocumentLoad Function """
    self.addGlobalFnc("NavBar(htmlObj, data)", ''' htmlObj.empty() ; data.forEach(function(rec){ htmlObj.append(rec) ; });  ''')

  def jsToUrl(self): return ''

  def __str__(self):
    """ Returns the HTML representation of the parameter bar """
    # Change the margin of the body as the Parameter bar will take some extra space on the top of the page
    self._report.jsOnLoadFnc.add('$("body").css("margin-bottom", "%spx")' % (self._report.marginTop - self.height))
    self._report.jsOnLoadFnc.add('''
      $('#nav_bar_reset').on('click', function(event) {%(jsPost)s});
      ''' % {'jsPost': self._report.jsPost("%s/locals/clean/%s" % (self._report._urlsApp['transfer'], self._report.run.report_name), None, self.jsGoTo(self.jsToUrlReset(), isPyData=False), isPyData=False)} )
    self._report.style.cssCls("CssParamsBar")
    self.addGlobalFnc("BreadCrumbClick(url)", '''
        $('#popup_loading').find("div").html("Loading "+ url +"...");
        $('#popup_loading_back').show(); $('#popup_loading').show();
        var params = []; console.log(%(breadCrumVar)s);
        for(var key in %(breadCrumVar)s['params']) {
          if (%(breadCrumVar)s['params'][key] != null){params.push(key +"="+ %(breadCrumVar)s['params'][key])}}
        if(params.length > 0) {breadCrumResult = url +"?"+ params.join("&")} else {breadCrumResult = url};
        window.location.href = breadCrumResult ''' % {'breadCrumVar': self._report.jsGlobal.breadCrumVar})
    return '''
      <div id="param_bar" class="%(clsParamBar)s" style="bottom:%(bottom)spx;border-bottom:1px solid %(border)s;z-index:30">
        <i class="fas fa-play" onclick="BreadCrumbClick(%(url)s)" style="margin-right:30px;cursor:pointer"></i><div %(strAttr)s></div>
      </div> ''' % {'bottom': 0, 'clsParamBar': self._report.style.cssName("CssParamsBar"),
                    'strAttr': self.get_attrs(pyClassNames=['CssBasicList']), "border": self._report.theme.greys[5], "url": self.jsToUrlReset()}


class HtmlSideBar(Html.Html):
  name, category, callFnc = 'Side bar', 'Others', 'sidebar'
  __reqCss, __reqJs = ['bootstrap', 'font-awesome', 'jquery-scrollbar'], ['bootstrap', 'font-awesome', 'jquery-scrollbar']
  onMouseOverColor = 'black'
  # _grpCls = GrpCls.CssClassSideBarFixed

  def __init__(self, report, links, color, size, servers):
    super(HtmlSideBar, self).__init__(report, links)
    self.color = self._report.theme.greys[0] if color is None else color
    self.size, self.servers = "%s%s" % (size[0], size[1]), servers
    self.css({'color': self.color, 'font-size': self.size, 'z-index': 5, 'margin': 0})
    self._actions = []

  @property
  def htmlId(self):
    return 'report_side_bar'

  def onDocumentLoadFnc(self):
    self.addGlobalFnc("%s(htmlObj, data)" % self.__class__.__name__, ''' $('#side_bar_explorer').empty();
      var group = $('<ul style="padding:0;margin:0;list-style:none"></ul>'); 
      var sidebarState = JSON.parse(window.sessionStorage.getItem('SIDEBAR_STATE'));
      if (sidebarState === null) {sidebarState = {}};
      data.folders.forEach(function(rec, index){
        var li = $('<li onclick="GetFolder(this);"></li>'); var display = 'none' ; var chevron = 'fas fa-folder';
        if ('env_' + index in sidebarState) { display = sidebarState['env_' + index] ; }
        if (display != 'none') { chevron = 'fas fa-folder-open'; }
        li.append('<a href="#" style="color:%(darkBlue)s;font-weight:bold;text-decoration:none;display:inline-block;padding-left:5px;margin-right:5px;font-size:14px;font-variant:small-caps;"><b><i id="chevron" class="fa ' + chevron + '"></i>&nbsp;'+ rec.name +'</a></b>') ;
        var sub = $('<ul id="env_' + index + '" style="padding:0 0 0 20px;margin:0;list-style:none;display:'+ display +'"></ul>') ;
        eventLi = $('<li onclick="event.stopPropagation();" style="margin-bottom:10px">') ;
        sub.append(eventLi);

        rec.items.forEach(function(data){
          var icons = '<i onclick=\\'GoToReport("%(urlReport)s/run/'+ rec.name +'/'+ data.script +'", true, false)\\' style="display:inline-block;margin-right:10px" title="Open in new tab" class="far fa-arrow-alt-circle-right"></i><i onclick=\\'GoToReport("%(urlReport)s/view/'+ rec.name +'/'+ data.script +'", true, false)\\' style="display:inline-block;font-size:14px;margin-right:10px" title="View Statistics" class="far fa-chart-bar"></i>' ;
          if (data.color == undefined) { data.color = '%(darkBlue)s' ; };
          if (data.doc) { icons = icons + '<i onclick=\\'GoToReport("%(urlReport)s/dsc/'+ rec.name +'/'+ data.script +'", true, false)\\' style="display:inline-block;margin-right:10px" title="Documentation" class="far fa-file-alt"></i>' ; }
          if (data.inFavorite) {sub.append('<li onclick="event.stopPropagation();"><a href="#" title="'+ data.script + '" style="display:inline-block;margin-right:10px;color:'+ data.color + '" onclick="BreadCrumbClick(this, \\''+ data.url.replace(/\\\\/g, '\\\\\\\\')  +'\\')">'+ data.name +'</a></li>') ; } 
          else { sub.append('<li style="white-space:nowrap; overflow-x: hidden;margin-right:10px" onclick="event.stopPropagation();"><a href="#" title="'+ data.script + '" style="display:inline-block;margin-right:10px;color:'+ data.color + '" onclick="BreadCrumbClick(this, \\''+ data.url.replace(/\\\\/g, '\\\\\\\\')  +'\\')">'+ data.name +'</a><i style="color:%(darkBlue)s;cursor:pointer;font-size:14px;margin-right:10px" onclick="Favorites(this, \\'' + rec.name + '\\', \\'' + data.script + '\\', \\'div\\', true)" title="Add to favorites" class="far fa-heart"></i></li>') ; }});
        li.append(sub); group.append(li); });

      $('#side_bar_explorer').append(group);
      $('#side_bar_explorer').css( {'height': parseInt($('#side_bar_envs').css('height'),10) - 250 + 'px'} ) ;
      if (data.favorites.length > 0){
        htmlObj.find('#side_bar_favorite').empty();
        var groupFav = $('<ul style="padding:0;margin:0;list-style:none"></ul>');
        data.favorites.forEach(function(rec, index){
          var li = $('<li onclick="GetFolder(this);"></li>'); var display = 'none' ; var chevron = 'fas fa-folder';
          if ('env_' + index in sidebarState) { display = sidebarState['env_' + index]}
          if (display != 'none') { chevron = 'fas fa-folder-open'}
          li.append('<a href="#" style="color:%(darkBlue)s;font-weight:bold;font-variant:small-caps;text-decoration:none;display:block;font-size:14px;padding-left:5px"><b><i id="chevron" class="fa ' + chevron + '"></i>&nbsp;'+ rec.name +'</a></b>') ;
          var sub = $('<ul id="env_' + index + '" style="padding:0 0 0 20px;margin:0;list-style:none;display:'+ display +'"></ul>') ;
          rec.items.forEach(function(data){
            if (data.color == undefined) { data.color = '%(darkBlue)s'};
            sub.append('<li onclick="event.stopPropagation();"><a href="#" title="'+ data.script + '" style="color:'+ data.color + '" onclick="BreadCrumbClick(this, \\''+ data.url.replace(/\\\\/g, '\\\\\\\\')  +'\\')">'+ data.name +'</a><div style="color:%(darkBlue)s;cursor:pointer;margin-left:5px;font-size:10px" onclick="Favorites(this, \\'' + rec.name + '\\', \\'' + data.script + '\\', \\'div\\', false)" title="Remove from favorites" class="fas fa-times"></div></li>') ; 
          });
          li.append(sub); groupFav.append(li)});
        $('#side_bar_favorite').append(groupFav)}
      $("#nav_link_title").tooltip()
      ''' % {'reportName': self._report.run.report_name, 'hoverColor': self.onMouseOverColor,
             'urlReport': "",
             'cssLinks': self._report.style.getClsTag(['CssSideBarLinks']), 'darkBlue': self._report.theme.colors[9]})

  def addAction(self, icon, url, tooltip='', color=None, isPyData=True):
    actionDef = ['<div style="position:relative;cursor:pointer;display:block;pointer-events:auto;margin-bottom:10px;">']
    if isPyData:
      actionDef.append(
        '<i onclick="GoToReport(\'%(url)s\')" class="%(icon)s" title="%(tooltip)s" style="color:%(color)s"></i>' % {
          "url": url, "icon": icon, "url": url, "color": color, "tooltip": tooltip})
    else:
      actionDef.append(
        '<i onclick="%(url)s" class="%(icon)s" title="%(tooltip)s" style="color:%(color)s"></i>' % {"url": url,
                                                                                                    "icon": icon,
                                                                                                    "url": url,
                                                                                                    "color": color,
                                                                                                    "tooltip": tooltip})
    actionDef.append('</div>')
    self._actions.append("".join(actionDef))
    return self

  def __str__(self):
    styleBarBulble, StyleBarMenu = "CssSideBarBubble", "CssSideBarMenu"
    # Add some specific classes for the sidebar items
    self._report.style.cssCls(styleBarBulble)
    self._report.style.cssCls(StyleBarMenu)
    return '''
      <div id="side_bar_envs" style='height:100%%;display:none;position:fixed;z-index:5;width:300px;top:0;background-color:%(lightBlue)s;left:0;padding:50px 5px 5px 45px;'>
        <span style="font-weight:bold;display:inline-block;text-transform:uppercase;margin-top:10px;width:100%%">Favorites</span>
        <div id="side_bar_favorite" style="width:100%%;display:inline-block"></div>

        <hr />
        <input type="text" style="width:100%%;display:inline-block" placeholder="search a report" />
        <span style="font-weight:bold;display:inline-block;text-transform:uppercase;margin-top:10px;width:100%%">Environments</span>
        <div id="side_bar_explorer" style="margin-right:20px;width:100%%;overflow-y:scroll"></div>
      </div>

      <div %(strAttr)s>                
        <div style="position:relative;cursor:pointer;display:inline-block;pointer-events:auto;" onclick="$('#side_bar_envs').toggle()">
          <i class="fas fa-users" title="Public folders"></i>
        </div>

        <hr style="background-color:white;margin:10px 5px 15px 5px" />
        %(actions)s

        <div style="position:fixed;bottom:5px;cursor:pointer;width:40px">
          <div style="position:relative;display:inline-block;pointer-events:auto;" title="Python Info" onmouseout="$(this).find('div').hide()" onmouseover="$(this).find('div').show();">
            <i onclick="GoToReport('%(urlReport)s/python', false, false)" style="margin-right:2px" class="fab fa-python"></i>
          </div>
          <div style="position:relative;display:inline-block;pointer-events:auto;" title="Settings" onmouseout="$(this).find('div').hide()" onmouseover="$(this).find('div').show();">
            <i onclick="GoToReport('%(urlAdmin)s/account', false, false)" style="margin-right:2px" class="fas fa-wrench"></i>
          </div>

          <br />
          <!--
          <div style="position:relative;display:inline-block;pointer-events:auto;" title="The Script scheduler" onmouseout="$(this).find('div').hide()" onmouseover="$(this).find('div').show();">
            <i onclick="GoToReport('%(urlReport)s/run/scheduler/index', false, false)"  class="far fa-clock"></i>
          </div> 
          <div style="position:relative;display:inline-block;pointer-events:auto;" title="The lab" onmouseout="$(this).find('div').hide()" onmouseover="$(this).find('div').show();">
            <i onclick="GoToReport('%(urlReport)s/run/lab/index', false, false)"  class="fas fa-flask"></i>
          </div> --!>
          <div style="position:relative;display:inline-block;pointer-events:auto;" title="Log off" onmouseout="$(this).find('div').hide()" onmouseover="$(this).find('div').show();">
            <i onclick="GoToReport('%(urlAdmin)s/logout', false, false)" style="margin-right:2px" class="fas fa-power-off"></i>
          </div>

        </div>
      </div>''' % {"strAttr": self.get_attrs(pyClassNames=self.defined),
                   'report_name': self._report.run.report_name, 'urlReport': "",
                   'cssDef': self._report.style.cssName(styleBarBulble),
                   'cssSideBarMenu': "",
                   'script_name': self._report.run.script_name,
                   'urlAdmin': "", 'lightBlue': self._report.theme.colors[1],
                   "actions": "".join(self._actions)}


class HtmlFooter(Html.Html):

  def __init__(self, report, components, width, height, profile):
    super(HtmlFooter, self).__init__(report, [], css_attrs={"width": width, "height": height}, profile=profile)
    self.__col_lst = None
    if components is not None:
      if not isinstance(components, list):
        components = [components]
      for c in components:
        self.__add__(c)

  @property
  def sections(self):
    if not self.__col_lst:
      self.__col_lst = []
    return self.__col_lst

  @sections.setter
  def sections(self, col_lst):
    self.__col_lst = col_lst

  @property
  def style(self):
    """
    Description:
    -----------
    Property to the CSS Style of the component

    :rtype: GrpClsMenu.ClassFooter
    """
    if self._styleObj is None:
      self._styleObj = GrpClsMenu.ClassFooter(self)
    return self._styleObj

  def __add__(self, htmlObj):
    """ Add items to the footer """
    htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    self.val.append(htmlObj)
    return self

  def __getitem__(self, i):
    """
    Return the internal column in the row for the given index

    :param i: the column index
    """
    return self.val[i]

  def add_menu(self):
    pass

  def __str__(self):
    str_h = "".join([val.html() for val in self.val])
    return "<footer %s>%s</footer>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_h)



"""
The location object contains information about the current URL.

The location object is part of the window object and is accessed through the window.location property.

Related Pages:

			https//www.w3schools.com/jsref/obj_location.asp

"""

from epyk.core.js import JsUtils

# All the predefined variable types
from epyk.core.js.fncs import JsFncs
from epyk.core.js.primitives import JsObject
from epyk.core.js.primitives import JsString


class JsLocation(object):
  @property
  def hostname(self):
    """
    The hostname property sets or returns the hostname of a URL.

    Example
    rptObj.location.hostname

    Related Pages:

			https//www.w3schools.com/jsref/obj_location.asp

    :return: Return the hostname property
    """
    return JsString.JsString("location.hostname", isPyData=False)

  @property
  def pathname(self):
    """
    The hostname property sets or returns the hostname of a URL.

    Example
    jsObj.location.pathname

    Related Pages:

			https//www.w3schools.com/jsref/obj_location.asp

    :return: Return the pathname property
    """
    return JsString.JsString("location.pathname", isPyData=False)

  @property
  def host(self):
    """
    The host property sets or returns the hostname and port of a URL.

    Example
    jsObj.location.host

    Related Pages:

			https//www.w3schools.com/jsref/prop_loc_host.asp

    :return: Return the hostname and port of the current URL
    """
    return JsString.JsString("location.host", isPyData=False)

  @property
  def hash(self):
    """
    The hash property sets or returns the anchor part of a URL, including the hash sign (#)

    Example
    jsObj.location.hash

    Related Pages:

			https//www.w3schools.com/jsref/prop_loc_hash.asp

    :return: A String, representing the anchor part of the URL, including the hash sign (#)
    """
    return JsObject.JsObject("location.hash", isPyData=False)

  @property
  def search(self):
    """
    The search property sets or returns the querystring part of a URL, including the question mark (?).

    Related Pages:

			https//www.w3schools.com/jsref/prop_loc_search.asp

    :return: A String, representing the querystring part of a URL, including the question mark (?)
    """
    return JsString.JsString("location.search", isPyData=False)

  @property
  def port(self):
    """
    The port property sets or returns the port number the server uses for a URL.

    Related Pages:

			https//www.w3schools.com/jsref/prop_loc_port.asp

    :return: A String, representing the port number of a URL.
    """
    return JsString.JsString("location.port", isPyData=False)

  @property
  def origin(self):
    """
    The origin property returns the protocol, hostname and port number of a URL.

    Example
    rptObj.js.location.origin + rptObj.js.location.pathname

    Related Pages:

			https//www.w3schools.com/jsref/prop_loc_origin.asp

    :return: A String, representing the protocol (including ://), the domain name (or IP address) and port number (including the colon sign (:) of the URL.
             For URL's using the "file:" protocol, the return value differs between browser
    """
    return JsString.JsString("location.origin", isPyData=False)

  def href(self, href=None, secured=False):
    """
    The href property sets or returns the entire URL of the current p

    Example
    rptObj.js.location.href("https://www.w3schools.com/howto/howto_js_fullscreen.asp")

    Related Pages:

			https://www.w3schools.com/jsref/prop_loc_href.asp

    :param href: Set the href property

    :return: A String, representing the entire URL of the page, including the protocol (like http://)
    """
    if href is None:
      return JsString.JsString("location.href", isPyData=False)

    if href.startswith("www."):
      href = r"http:\\%s" % href if not secured else r"https:\\%s" % href
    return JsObject.JsObject("location.href = %s" % JsUtils.jsConvertData(href, None))

  def open_new_tab(self, url, name="_blank", specs=None, replace=None, windowId="window", secured=False):
    """
    Opens a new browser window in a new tab (duplicated but part of the Window module)

    Example
    rptObj.js.location.open_new_tab("www.google.fr")

    Related Pages:

			https://www.w3schools.com/Jsref/met_win_open.asp

    :param url: Optional. Specifies the URL of the page to open. If no URL is specified, a new window/tab with about:blank is opened
    :param name: Optional. Specifies the target attribute or the name of the window. Default _blank
    :param specs: Optional. A comma-separated list of items, no whitespaces.
    :param replace: Optional. Specifies whether the URL creates a new entry or replaces the current entry in the history list
    :param windowId: The JavaScript window object

    :return:
    """
    if url.startswith("www."):
      url = r"http:\\%s" % url if not secured else r"https:\\%s" % url
    url = JsUtils.jsConvertData(url, None)
    name = JsUtils.jsConvertData(name, None)
    specs = JsUtils.jsConvertData(specs, None)
    replace = JsUtils.jsConvertData(replace, None)
    return JsFncs.JsFunction("%s.open(%s, %s, %s, %s)" % (windowId, url, name, specs, replace))

  def mail(self, mails, subject, body):
    """
    The mailto link when clicked opens users default email program or software. A new email page is created with "To" field containing the address of the name specified on the link by default.

    Example
    rptObj.js.location.mail(["test@gmail.com"], "This is a test", "This is the email's content")

    Documentation
    http://www.tutorialspark.com/html5/HTML5_email_mailto.php

    :param mails: A list of email addresses
    :param subject: The email's subject
    :param body: The email's content

    :return: THe Javascript string
    """
    if isinstance(mails, list):
      mails = ";".join(mails)
    return self.href(JsString.JsString("'mailto:%s?subject=%s&body='+ encodeURIComponent('%s')" % (mails, subject, body), isPyData=False))

  def reload(self, forceGet=False):
    """
    The reload() method is used to reload the current document.

    The reload() method does the same as the reload button in your browser.

    Related Pages:

			https//www.w3schools.com/jsref/met_loc_reload.asp

    :param forceGet: Optional. Specifies the type of reloading:
          false - Default. Reloads the current page from the cache.
          true - Reloads the current page from the server

    :return: Void
    """
    forceGet = JsUtils.jsConvertData(forceGet, None)
    return JsFncs.JsFunction("location.reload(%s)" % forceGet)

  def assign(self, url):
    """
    The assign() method loads a new document.

    Related Pages:

			https//www.w3schools.com/jsref/met_loc_assign.asp

    :param url: Required. Specifies the URL of the page to navigate to

    :return: Void
    """
    jsData = JsUtils.jsConvertData(url, None)
    return JsFncs.JsFunction("location.assign(%s)" % jsData)

  def replace(self, url, secured=False):
    """
    The replace() method replaces the current document with a new one.

    The difference between this method and assign(), is that replace() removes the URL of the current document from the document history, meaning that it is not possible to use the "back" button to navigate back to the original document.

    Related Pages:

			https//www.w3schools.com/jsref/met_loc_replace.asp

    :param url: Required. Specifies the URL of the page to navigate to
    :param secured: Optional. If the http is missing. This will be used to fix the url

    :return: Void
    """
    if url.startswith("www."):
      url = r"http:\\%s" % url if not secured else r"https:\\%s" % url
    js_data = JsUtils.jsConvertData(url, None)
    return JsFncs.JsFunction("location.replace(%s)" % js_data)

  def postTo(self, url, data, method="POST", target="_blank"):
    """
    This method will create a internal form and submit the response exactly like a post of a form to another page

    Related Pages:

			https://www.w3schools.com/jsref/dom_obj_form.asp

    :param url: The target url
    :param data: A python dictionary
    :param method: Optional. The method used to send the data. Default POST

    """
    inputs = []
    for k, v in data.items():
      jsVal = JsUtils.jsConvertData(v, None)
      inputs.append('var input = document.createElement("input"); input.name = "%s"; input.type = "hidden"; input.value = %s; form.appendChild(input);' % (k, jsVal))
    return ''' 
      var form = document.createElement("form"); form.method = "%s"; form.target = "%s"; form.action = "%s"; %s;
      document.body.appendChild(form); form.submit()''' % (method, target, url, "".join(inputs))
